# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import six

from oslo_log import log
from oslo_utils import excutils
from oslo_utils import importutils

storops = importutils.try_import('storops')
if storops:
    # pylint: disable=import-error
    from storops import exception as storops_ex
    from storops.unity import enums

from manila.common import constants as const
from manila import exception
from manila.i18n import _
from manila.share.drivers.dell_emc.common.enas import utils as enas_utils
from manila.share.drivers.dell_emc.plugins.unity import utils

LOG = log.getLogger(__name__)


class UnityClient(object):
    def __init__(self, host, username, password):
        if storops is None:
            LOG.error('StorOps is required to run EMC Unity driver.')
        self.system = storops.UnitySystem(host, username, password)

    def create_cifs_share(self, resource, share_name):
        """Create CIFS share from the resource.

        :param resource: either UnityFilesystem or UnitySnap object
        :param share_name: CIFS share name
        :return: UnityCifsShare object
        """
        try:
            return resource.create_cifs_share(share_name)
        except storops_ex.UnitySmbShareNameExistedError:
            return self.get_share(share_name, 'CIFS')

    def create_nfs_share(self, resource, share_name):
        """Create NFS share from the resource.

        :param resource: either UnityFilesystem or UnitySnap object
        :param share_name: NFS share name
        :return: UnityNfsShare object
        """
        try:
            return resource.create_nfs_share(share_name)
        except storops_ex.UnityNfsShareNameExistedError:
            return self.get_share(share_name, 'NFS')

    def create_nfs_filesystem_and_share(self, pool, nas_server,
                                        share_name, size_gb):
        """Create filesystem and share from pool/NAS server.

        :param pool: pool for file system creation
        :param nas_server: nas server for file system creation
        :param share_name: file system and share name
        :param size_gb: file system size
        """
        size = utils.gib_to_byte(size_gb)
        pool.create_nfs_share(
            nas_server, share_name, size, user_cap=True)

    def get_share(self, name, share_proto):
        # Validate the share protocol
        proto = share_proto.upper()

        if proto == 'CIFS':
            return self.system.get_cifs_share(name=name)
        elif proto == 'NFS':
            return self.system.get_nfs_share(name=name)
        else:
            raise exception.BadConfigurationException(
                reason=_('Invalid NAS protocol supplied: %s.') % share_proto)

    @staticmethod
    def delete_share(share):
        share.delete()

    def create_filesystem(self, pool, nas_server, share_name, size_gb, proto):
        try:
            size = utils.gib_to_byte(size_gb)
            return pool.create_filesystem(nas_server,
                                          share_name,
                                          size,
                                          proto=proto,
                                          user_cap=True)
        except storops_ex.UnityFileSystemNameAlreadyExisted:
            LOG.debug('Filesystem %s already exists, '
                      'ignoring filesystem creation.', share_name)
            return self.system.get_filesystem(name=share_name)

    @staticmethod
    def delete_filesystem(filesystem):
        try:
            filesystem.delete()
        except storops_ex.UnityResourceNotFoundError:
            LOG.info('Filesystem %s is already removed.', filesystem.name)

    def create_nas_server(self, name, sp, pool, tenant=None):
        try:
            return self.system.create_nas_server(name, sp, pool,
                                                 tenant=tenant)
        except storops_ex.UnityNasServerNameUsedError:
            LOG.info('Share server %s already exists, ignoring share '
                     'server creation.', name)
            return self.get_nas_server(name)

    def get_nas_server(self, name):
        try:
            return self.system.get_nas_server(name=name)
        except storops_ex.UnityResourceNotFoundError:
            LOG.info('NAS server %s not found.', name)
            raise

    def delete_nas_server(self, name, username=None, password=None):
        tenant = None
        try:
            nas_server = self.get_nas_server(name=name)
            tenant = nas_server.tenant
            nas_server.delete(username=username, password=password)
        except storops_ex.UnityResourceNotFoundError:
            LOG.info('NAS server %s not found.', name)

        if tenant is not None:
            self._delete_tenant(tenant)

    @staticmethod
    def _delete_tenant(tenant):
        if tenant.nas_servers:
            LOG.debug('There are NAS servers belonging to the tenant %s. '
                      'Do not delete it.',
                      tenant.get_id())
            return
        try:
            tenant.delete(delete_hosts=True)
        except storops_ex.UnityException as ex:
            LOG.warning('Delete tenant %(tenant)s failed with error: '
                        '%(ex)s. Leave the tenant on the system.',
                        {'tenant': tenant.get_id(),
                         'ex': ex})

    @staticmethod
    def create_dns_server(nas_server, domain, dns_ip):
        try:
            nas_server.create_dns_server(domain, dns_ip)
        except storops_ex.UnityOneDnsPerNasServerError:
            LOG.info('DNS server %s already exists, '
                     'ignoring DNS server creation.', domain)

    @staticmethod
    def create_interface(nas_server, ip_addr, netmask, gateway, port_id,
                         vlan_id=None, prefix_length=None):
        try:
            nas_server.create_file_interface(port_id,
                                             ip_addr,
                                             netmask=netmask,
                                             v6_prefix_length=prefix_length,
                                             gateway=gateway,
                                             vlan_id=vlan_id)
        except storops_ex.UnityIpAddressUsedError:
            raise exception.IPAddressInUse(ip=ip_addr)

    @staticmethod
    def enable_cifs_service(nas_server, domain, username, password):
        try:
            nas_server.enable_cifs_service(
                nas_server.file_interface,
                domain=domain,
                domain_username=username,
                domain_password=password)
        except storops_ex.UnitySmbNameInUseError:
            LOG.info('CIFS service on NAS server %s is '
                     'already enabled.', nas_server.name)

    @staticmethod
    def enable_nfs_service(nas_server):
        try:
            nas_server.enable_nfs_service()
        except storops_ex.UnityNfsAlreadyEnabledError:
            LOG.info('NFS service on NAS server %s is '
                     'already enabled.', nas_server.name)

    @staticmethod
    def create_snapshot(filesystem, name):
        access_type = enums.FilesystemSnapAccessTypeEnum.CHECKPOINT
        try:
            return filesystem.create_snap(name, fs_access_type=access_type)
        except storops_ex.UnitySnapNameInUseError:
            LOG.info('Snapshot %(snap)s on Filesystem %(fs)s already '
                     'exists.', {'snap': name, 'fs': filesystem.name})

    def create_snap_of_snap(self, src_snap, dst_snap_name):
        if isinstance(src_snap, six.string_types):
            snap = self.get_snapshot(name=src_snap)
        else:
            snap = src_snap

        try:
            return snap.create_snap(dst_snap_name)
        except storops_ex.UnitySnapNameInUseError:
            return self.get_snapshot(dst_snap_name)

    def get_snapshot(self, name):
        return self.system.get_snap(name=name)

    @staticmethod
    def delete_snapshot(snap):
        try:
            snap.delete()
        except storops_ex.UnityResourceNotFoundError:
            LOG.info('Snapshot %s is already removed.', snap.name)

    def get_pool(self, name=None):
        return self.system.get_pool(name=name)

    def get_storage_processor(self, sp_id=None):
        sp = self.system.get_sp(sp_id)
        if sp_id is None:
            # `sp` is a list of SPA and SPB.
            return [s for s in sp if s is not None and s.existed]
        else:
            return sp if sp.existed else None

    def cifs_clear_access(self, share_name, white_list=None):
        share = self.system.get_cifs_share(name=share_name)
        share.clear_access(white_list)

    def nfs_clear_access(self, share_name, white_list=None):
        share = self.system.get_nfs_share(name=share_name)
        share.clear_access(white_list, force_create_host=True)

    def cifs_allow_access(self, share_name, user_name, access_level):
        share = self.system.get_cifs_share(name=share_name)

        if access_level == const.ACCESS_LEVEL_RW:
            cifs_access = enums.ACEAccessLevelEnum.WRITE
        else:
            cifs_access = enums.ACEAccessLevelEnum.READ

        share.add_ace(user=user_name, access_level=cifs_access)

    def nfs_allow_access(self, share_name, host_ip, access_level):
        share = self.system.get_nfs_share(name=share_name)
        host_ip = enas_utils.convert_ipv6_format_if_needed(host_ip)
        if access_level == const.ACCESS_LEVEL_RW:
            share.allow_read_write_access(host_ip, force_create_host=True)
            share.allow_root_access(host_ip, force_create_host=True)
        else:
            share.allow_read_only_access(host_ip, force_create_host=True)

    def cifs_deny_access(self, share_name, user_name):
        share = self.system.get_cifs_share(name=share_name)

        try:
            share.delete_ace(user=user_name)
        except storops_ex.UnityAclUserNotFoundError:
            LOG.debug('ACL User "%(user)s" does not exist.',
                      {'user': user_name})

    def nfs_deny_access(self, share_name, host_ip):
        share = self.system.get_nfs_share(name=share_name)

        try:
            share.delete_access(host_ip)
        except storops_ex.UnityHostNotFoundException:
            LOG.info('%(host)s access to %(share)s is already removed.',
                     {'host': host_ip, 'share': share_name})

    def get_file_ports(self):
        ports = self.system.get_file_port()
        link_up_ports = []
        for port in ports:
            if port.is_link_up and self._is_external_port(port.id):
                link_up_ports.append(port)

        return link_up_ports

    def extend_filesystem(self, fs, new_size_gb):
        size = utils.gib_to_byte(new_size_gb)
        try:
            fs.extend(size, user_cap=True)
        except storops_ex.UnityNothingToModifyError:
            LOG.debug('The size of the file system %(id)s is %(size)s '
                      'bytes.', {'id': fs.get_id(), 'size': size})
        return size

    def shrink_filesystem(self, share_id, fs, new_size_gb):
        size = utils.gib_to_byte(new_size_gb)
        try:
            fs.shrink(size, user_cap=True)
        except storops_ex.UnityNothingToModifyError:
            LOG.debug('The size of the file system %(id)s is %(size)s '
                      'bytes.', {'id': fs.get_id(), 'size': size})
        except storops_ex.UnityShareShrinkSizeTooSmallError:
            LOG.error('The used size of the file system %(id)s is '
                      'bigger than input shrink size,'
                      'it may cause date loss.', {'id': fs.get_id()})
            raise exception.ShareShrinkingPossibleDataLoss(share_id=share_id)
        return size

    @staticmethod
    def _is_external_port(port_id):
        return 'eth' in port_id or '_la' in port_id

    def get_tenant(self, name, vlan_id):
        if not vlan_id:
            # Do not create vlan for flat network
            return None

        tenant = None
        try:
            tenant_name = "vlan_%(vlan_id)s_%(name)s" % {'vlan_id': vlan_id,
                                                         'name': name}
            tenant = self.system.create_tenant(tenant_name, vlans=[vlan_id])
        except (storops_ex.UnityVLANUsedByOtherTenantError,
                storops_ex.UnityTenantNameInUseError,
                storops_ex.UnityVLANAlreadyHasInterfaceError):
            with excutils.save_and_reraise_exception() as exc:
                tenant = self.system.get_tenant_use_vlan(vlan_id)
                if tenant is not None:
                    LOG.debug("The VLAN %s is already added into a tenant. "
                              "Use the existing VLAN tenant.", vlan_id)
                    exc.reraise = False
        except storops_ex.SystemAPINotSupported:
            LOG.info("This system doesn't support tenant.")

        return tenant

    def restore_snapshot(self, snap_name):
        snap = self.get_snapshot(snap_name)
        return snap.restore(delete_backup=True)



# Copyright 2016 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Module for storing ZFSonLinux driver utility stuff such as:
 - Common ZFS code
 - Share helpers
"""

# TODO(vponomaryov): add support of SaMBa

import abc

from oslo_log import log
import six

from manila.common import constants
from manila import exception
from manila.i18n import _
from manila.share import driver
from manila.share.drivers.ganesha import utils as ganesha_utils
from manila import utils

LOG = log.getLogger(__name__)


def zfs_dataset_synchronized(f):

    def wrapped_func(self, *args, **kwargs):
        key = "zfs-dataset-%s" % args[0]

        @utils.synchronized(key, external=True)
        def source_func(self, *args, **kwargs):
            return f(self, *args, **kwargs)

        return source_func(self, *args, **kwargs)

    return wrapped_func


def get_remote_shell_executor(
        ip, port, conn_timeout, login=None, password=None, privatekey=None,
        max_size=10):
    return ganesha_utils.SSHExecutor(
        ip=ip,
        port=port,
        conn_timeout=conn_timeout,
        login=login,
        password=password,
        privatekey=privatekey,
        max_size=max_size,
    )


class ExecuteMixin(driver.ExecuteMixin):

    def init_execute_mixin(self, *args, **kwargs):
        """Init method for mixin called in the end of driver's __init__()."""
        super(ExecuteMixin, self).init_execute_mixin(*args, **kwargs)
        if self.configuration.zfs_use_ssh:
            self.ssh_executor = get_remote_shell_executor(
                ip=self.configuration.zfs_service_ip,
                port=22,
                conn_timeout=self.configuration.ssh_conn_timeout,
                login=self.configuration.zfs_ssh_username,
                password=self.configuration.zfs_ssh_user_password,
                privatekey=self.configuration.zfs_ssh_private_key_path,
                max_size=10,
            )
        else:
            self.ssh_executor = None

    def execute(self, *cmd, **kwargs):
        """Common interface for running shell commands."""
        if kwargs.get('executor'):
            executor = kwargs.get('executor')
        elif self.ssh_executor:
            executor = self.ssh_executor
        else:
            executor = self._execute
        kwargs.pop('executor', None)
        if cmd[0] == 'sudo':
            kwargs['run_as_root'] = True
            cmd = cmd[1:]
        return executor(*cmd, **kwargs)

    @utils.retry(retry_param=exception.ProcessExecutionError,
                 interval=5, retries=36, backoff_rate=1)
    def execute_with_retry(self, *cmd, **kwargs):
        """Retry wrapper over common shell interface."""
        try:
            return self.execute(*cmd, **kwargs)
        except exception.ProcessExecutionError as e:
            LOG.warning("Failed to run command, got error: %s", e)
            raise

    def _get_option(self, resource_name, option_name, pool_level=False,
                    **kwargs):
        """Returns value of requested zpool or zfs dataset option."""
        app = 'zpool' if pool_level else 'zfs'

        out, err = self.execute(
            'sudo', app, 'get', option_name, resource_name, **kwargs)

        data = self.parse_zfs_answer(out)
        option = data[0]['VALUE']
        msg_payload = {'option': option_name, 'value': option}
        LOG.debug("ZFS option %(option)s's value is %(value)s.", msg_payload)
        return option

    def parse_zfs_answer(self, string):
        """Returns list of dicts with data returned by ZFS shell commands."""
        lines = string.split('\n')
        if len(lines) < 2:
            return []
        keys = list(filter(None, lines[0].split(' ')))
        data = []
        for line in lines[1:]:
            values = list(filter(None, line.split(' ')))
            if not values:
                continue
            data.append(dict(zip(keys, values)))
        return data

    def get_zpool_option(self, zpool_name, option_name, **kwargs):
        """Returns value of requested zpool option."""
        return self._get_option(zpool_name, option_name, True, **kwargs)

    def get_zfs_option(self, dataset_name, option_name, **kwargs):
        """Returns value of requested zfs dataset option."""
        return self._get_option(dataset_name, option_name, False, **kwargs)

    def zfs(self, *cmd, **kwargs):
        """ZFS shell commands executor."""
        return self.execute('sudo', 'zfs', *cmd, **kwargs)

    def zfs_with_retry(self, *cmd, **kwargs):
        """ZFS shell commands executor."""
        return self.execute_with_retry('sudo', 'zfs', *cmd, **kwargs)


@six.add_metaclass(abc.ABCMeta)
class NASHelperBase(object):
    """Base class for share helpers of 'ZFS on Linux' driver."""

    def __init__(self, configuration):
        """Init share helper.

        :param configuration: share driver 'configuration' instance
        :return: share helper instance.
        """
        self.configuration = configuration
        self.init_execute_mixin()  # pylint: disable=no-member
        self.verify_setup()

    @abc.abstractmethod
    def verify_setup(self):
        """Performs checks for required stuff."""

    @abc.abstractmethod
    def create_exports(self, dataset_name, executor):
        """Creates share exports."""

    @abc.abstractmethod
    def get_exports(self, dataset_name, service, executor):
        """Gets/reads share exports."""

    @abc.abstractmethod
    def remove_exports(self, dataset_name, executor):
        """Removes share exports."""

    @abc.abstractmethod
    def update_access(self, dataset_name, access_rules, add_rules,
                      delete_rules, executor):
        """Update access rules for specified ZFS dataset."""


class NFSviaZFSHelper(ExecuteMixin, NASHelperBase):
    """Helper class for handling ZFS datasets as NFS shares.

    Kernel and Fuse versions of ZFS have different syntax for setting up access
    rules, and this Helper designed to satisfy both making autodetection.
    """

    @property
    def is_kernel_version(self):
        """Says whether Kernel version of ZFS is used or not."""
        if not hasattr(self, '_is_kernel_version'):
            try:
                self.execute('modinfo', 'zfs')
                self._is_kernel_version = True
            except exception.ProcessExecutionError as e:
                LOG.info(
                    "Looks like ZFS kernel module is absent. "
                    "Assuming FUSE version is installed. Error: %s", e)
                self._is_kernel_version = False
        return self._is_kernel_version

    def verify_setup(self):
        """Performs checks for required stuff."""
        out, err = self.execute('which', 'exportfs')
        if not out:
            raise exception.ZFSonLinuxException(
                msg=_("Utility 'exportfs' is not installed."))
        try:
            self.execute('sudo', 'exportfs')
        except exception.ProcessExecutionError:
            LOG.exception("Call of 'exportfs' utility returned error.")
            raise

        # Init that class instance attribute on start of manila-share service
        self.is_kernel_version

    def create_exports(self, dataset_name, executor=None):
        """Creates NFS share exports for given ZFS dataset."""
        return self.get_exports(dataset_name, executor=executor)

    def get_exports(self, dataset_name, executor=None):
        """Gets/reads NFS share export for given ZFS dataset."""
        mountpoint = self.get_zfs_option(
            dataset_name, 'mountpoint', executor=executor)
        return [
            {
                "path": "%(ip)s:%(mp)s" % {"ip": ip, "mp": mountpoint},
                "metadata": {
                },
                "is_admin_only": is_admin_only,
            } for ip, is_admin_only in (
                (self.configuration.zfs_share_export_ip, False),
                (self.configuration.zfs_service_ip, True))
        ]

    @zfs_dataset_synchronized
    def remove_exports(self, dataset_name, executor=None):
        """Removes NFS share exports for given ZFS dataset."""
        sharenfs = self.get_zfs_option(
            dataset_name, 'sharenfs', executor=executor)
        if sharenfs == 'off':
            return
        self.zfs("set", "sharenfs=off", dataset_name, executor=executor)

    def _get_parsed_access_to(self, access_to):
        netmask = utils.cidr_to_netmask(access_to)
        if netmask == '255.255.255.255':
            return access_to.split('/')[0]
        return access_to.split('/')[0] + '/' + netmask

    @zfs_dataset_synchronized
    def update_access(self, dataset_name, access_rules, add_rules,
                      delete_rules, make_all_ro=False, executor=None):
        """Update access rules for given ZFS dataset exported as NFS share."""
        rw_rules = []
        ro_rules = []
        for rule in access_rules:
            if rule['access_type'].lower() != 'ip':
                msg = _("Only IP access type allowed for NFS protocol.")
                raise exception.InvalidShareAccess(reason=msg)
            if (rule['access_level'] == constants.ACCESS_LEVEL_RW and
                    not make_all_ro):
                rw_rules.append(self._get_parsed_access_to(rule['access_to']))
            elif (rule['access_level'] in (constants.ACCESS_LEVEL_RW,
                                           constants.ACCESS_LEVEL_RO)):
                ro_rules.append(self._get_parsed_access_to(rule['access_to']))
            else:
                msg = _("Unsupported access level provided - "
                        "%s.") % rule['access_level']
                raise exception.InvalidShareAccess(reason=msg)

        rules = []
        if self.is_kernel_version:
            if rw_rules:
                rules.append(
                    "rw=%s,no_root_squash" % ":".join(rw_rules))
            if ro_rules:
                rules.append("ro=%s,no_root_squash" % ":".join(ro_rules))
            rules_str = "sharenfs=" + (','.join(rules) or 'off')
        else:
            for rule in rw_rules:
                rules.append("%s:rw,no_root_squash" % rule)
            for rule in ro_rules:
                rules.append("%s:ro,no_root_squash" % rule)
            rules_str = "sharenfs=" + (' '.join(rules) or 'off')

        out, err = self.zfs(
            'list', '-r', dataset_name.split('/')[0], executor=executor)
        data = self.parse_zfs_answer(out)
        for datum in data:
            if datum['NAME'] == dataset_name:
                self.zfs("set", rules_str, dataset_name)
                break
        else:
            LOG.warning(
                "Dataset with '%(name)s' NAME is absent on backend. "
                "Access rules were not applied.", {'name': dataset_name})

        # NOTE(vponomaryov): Setting of ZFS share options does not remove rules
        # that were added and then removed. So, remove them explicitly.
        if delete_rules and access_rules:
            mountpoint = self.get_zfs_option(dataset_name, 'mountpoint')
            for rule in delete_rules:
                if rule['access_type'].lower() != 'ip':
                    continue
                access_to = self._get_parsed_access_to(rule['access_to'])
                export_location = access_to + ':' + mountpoint
                self.execute(
                    'sudo', 'exportfs', '-u', export_location,
                    executor=executor,
                )



# Copyright (c) 2021 NetApp, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unit tests for the Security Service helper module."""

from unittest import mock

import ddt

from manila import exception
from manila.share import configuration
from manila.share.drivers.container import security_service_helper
from manila import test
from manila.tests import db_utils


INVALID_CREDENTIALS_EXIT_CODE = 49


@ddt.ddt
class SecurityServiceHelperTestCase(test.TestCase):
    """Tests DockerExecHelper"""

    def setUp(self):
        super(SecurityServiceHelperTestCase, self).setUp()
        self.fake_conf = configuration.Configuration(None)
        self.fake_conf.container_image_name = "fake_image"
        self.fake_conf.container_volume_mount_path = "/tmp/shares"
        self.security_service_helper = (
            security_service_helper.SecurityServiceHelper(
                configuration=self.fake_conf))

    def test_setup_security_service(self):
        share_server = db_utils.create_share_server()
        security_service = db_utils.create_security_service()

        mock_ldap_bind = self.mock_object(
            self.security_service_helper, 'ldap_bind')

        self.security_service_helper.setup_security_service(
            share_server['id'], security_service)

        mock_ldap_bind.assert_called_once_with(
            share_server['id'], security_service)

    def test_update_security_service(self):
        share_server = db_utils.create_share_server()
        current_security_service = db_utils.create_security_service()
        new_security_service = db_utils.create_security_service()

        mock_ldap_bind = self.mock_object(
            self.security_service_helper, 'ldap_bind')

        self.security_service_helper.update_security_service(
            share_server['id'], current_security_service, new_security_service)

        mock_ldap_bind.assert_called_once_with(
            share_server['id'], new_security_service)

    def _setup_test_ldap_bind_tests(self):
        share_server = db_utils.create_security_service()
        security_service = db_utils.create_security_service()
        ldap_get_info = {
            'ss_password': security_service['password'],
            'ss_user': security_service['user']
        }
        expected_cmd = [
            "docker", "exec", "%s" % share_server['id'], "ldapwhoami", "-x",
            "-H", "ldap://localhost:389", "-D",
                              "cn=%s,dc=example,dc=com" % ldap_get_info[
                                  "ss_user"],
            "-w", "%s" % ldap_get_info["ss_password"]]

        return share_server, security_service, ldap_get_info, expected_cmd

    def test_ldap_bind(self):
        share_server, security_service, ldap_get_info, expected_cmd = (
            self._setup_test_ldap_bind_tests())

        mock_ldap_get_info = self.mock_object(
            self.security_service_helper, 'ldap_get_info',
            mock.Mock(return_value=ldap_get_info))
        mock_ldap_retry_operation = self.mock_object(
            self.security_service_helper, 'ldap_retry_operation')

        self.security_service_helper.ldap_bind(
            share_server['id'], security_service)

        mock_ldap_get_info.assert_called_once_with(security_service)
        mock_ldap_retry_operation.assert_called_once_with(expected_cmd,
                                                          run_as_root=True)

    def test_ldap_get_info(self):
        security_service = db_utils.create_security_service()
        expected_ldap_get_info = {
            'ss_password': security_service['password'],
            'ss_user': security_service['user']
        }

        ldap_get_info = self.security_service_helper.ldap_get_info(
            security_service)

        self.assertEqual(expected_ldap_get_info, ldap_get_info)

    @ddt.data(
        {'type': 'ldap'},
        {'user': 'fake_user'},
        {'password': 'fake_password'},
    )
    def test_ldap_get_info_exception(self, sec_service_data):
        self.assertRaises(
            exception.ShareBackendException,
            self.security_service_helper.ldap_get_info,
            sec_service_data
        )

    def test_ldap_retry_operation(self):
        mock_cmd = ["command", "to", "be", "executed"]

        mock_execute = self.mock_object(self.security_service_helper,
                                        '_execute')

        self.security_service_helper.ldap_retry_operation(mock_cmd,
                                                          run_as_root=True)

        mock_execute.assert_called_once_with(*mock_cmd, run_as_root=True)

    def test_ldap_retry_operation_timeout(self):
        mock_cmd = ["command", "to", "be", "executed"]

        mock_execute = self.mock_object(
            self.security_service_helper, '_execute',
            mock.Mock(
                side_effect=exception.ProcessExecutionError(exit_code=1)))

        self.assertRaises(
            exception.ShareBackendException,
            self.security_service_helper.ldap_retry_operation,
            mock_cmd,
            run_as_root=False,
            timeout=10)

        mock_execute.assert_has_calls([
            mock.call(*mock_cmd, run_as_root=False),
            mock.call(*mock_cmd, run_as_root=False)])

    def test_ldap_retry_operation_invalid_credential(self):
        mock_cmd = ["command", "to", "be", "executed"]

        mock_execute = self.mock_object(
            self.security_service_helper, '_execute',
            mock.Mock(
                side_effect=exception.ProcessExecutionError(
                    exit_code=49)))

        self.assertRaises(
            exception.ShareBackendException,
            self.security_service_helper.ldap_retry_operation,
            mock_cmd,
            run_as_root=False)

        mock_execute.assert_called_once_with(*mock_cmd, run_as_root=False)



# Copyright 2013 NetApp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manila.api import common


class ViewBuilder(common.ViewBuilder):
    """Model a server API response as a python dictionary."""

    _collection_name = 'snapshots'
    _detail_version_modifiers = [
        "add_provider_location_field",
        "add_project_and_user_ids",
    ]

    def summary_list(self, request, snapshots):
        """Show a list of share snapshots without many details."""
        return self._list_view(self.summary, request, snapshots)

    def detail_list(self, request, snapshots):
        """Detailed view of a list of share snapshots."""
        return self._list_view(self.detail, request, snapshots)

    def summary(self, request, snapshot):
        """Generic, non-detailed view of an share snapshot."""
        return {
            'snapshot': {
                'id': snapshot.get('id'),
                'name': snapshot.get('display_name'),
                'links': self._get_links(request, snapshot['id'])
            }
        }

    def detail(self, request, snapshot):
        """Detailed view of a single share snapshot."""
        snapshot_dict = {
            'id': snapshot.get('id'),
            'share_id': snapshot.get('share_id'),
            'share_size': snapshot.get('share_size'),
            'created_at': snapshot.get('created_at'),
            'status': snapshot.get('aggregate_status'),
            'name': snapshot.get('display_name'),
            'description': snapshot.get('display_description'),
            'size': snapshot.get('size'),
            'share_proto': snapshot.get('share_proto'),
            'links': self._get_links(request, snapshot['id']),
        }

        self.update_versioned_resource_dict(request, snapshot_dict, snapshot)

        return {'snapshot': snapshot_dict}

    @common.ViewBuilder.versioned_method("2.12")
    def add_provider_location_field(self, context, snapshot_dict, snapshot):
        # NOTE(xyang): Only retrieve provider_location for admin.
        if context.is_admin:
            snapshot_dict['provider_location'] = snapshot.get(
                'provider_location')

    @common.ViewBuilder.versioned_method("2.17")
    def add_project_and_user_ids(self, context, snapshot_dict, snapshot):
        snapshot_dict['user_id'] = snapshot.get('user_id')
        snapshot_dict['project_id'] = snapshot.get('project_id')

    def _list_view(self, func, request, snapshots):
        """Provide a view for a list of share snapshots."""
        snapshots_list = [func(request, snapshot)['snapshot']
                          for snapshot in snapshots]
        snapshots_links = self._get_collection_links(request,
                                                     snapshots,
                                                     self._collection_name)
        snapshots_dict = {self._collection_name: snapshots_list}

        if snapshots_links:
            snapshots_dict['share_snapshots_links'] = snapshots_links

        return snapshots_dict

    def detail_access(self, request, access):
        access = {
            'snapshot_access': {
                'id': access['id'],
                'access_type': access['access_type'],
                'access_to': access['access_to'],
                'state': access['state'],
            }
        }
        return access

    def detail_list_access(self, request, access_list):
        return {
            'snapshot_access_list':
                ([self.detail_access(request, access)['snapshot_access']
                 for access in access_list])
        }



# Copyright (c) 2016 Mirantis, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unit tests for the Container helper module."""

from unittest import mock
import uuid

import ddt

from manila import exception
from manila.share import configuration
from manila.share.drivers.container import container_helper
from manila import test
from manila.tests.share.drivers.container import fakes


@ddt.ddt
class DockerExecHelperTestCase(test.TestCase):
    """Tests DockerExecHelper"""

    def setUp(self):
        super(DockerExecHelperTestCase, self).setUp()
        self.fake_conf = configuration.Configuration(None)
        self.fake_conf.container_image_name = "fake_image"
        self.fake_conf.container_volume_mount_path = "/tmp/shares"
        self.DockerExecHelper = container_helper.DockerExecHelper(
            configuration=self.fake_conf)

    def test_start_container(self):
        self.mock_object(self.DockerExecHelper, "_inner_execute",
                         mock.Mock(return_value=['fake_output', '']))
        uuid.uuid1 = mock.Mock(return_value='')
        expected = ['docker', 'run', '-d', '-i', '-t', '--privileged', '-v',
                    '/dev:/dev', '--name=manila_cifs_docker_container', '-v',
                    '/tmp/shares:/shares', 'fake_image']

        self.DockerExecHelper.start_container()

        self.DockerExecHelper._inner_execute.assert_called_once_with(expected)

    def test_start_container_impossible_failure(self):
        self.mock_object(self.DockerExecHelper, "_inner_execute",
                         mock.Mock(side_effect=OSError()))

        self.assertRaises(exception.ShareBackendException,
                          self.DockerExecHelper.start_container)

    def test_stop_container(self):
        self.mock_object(self.DockerExecHelper, "_inner_execute",
                         mock.Mock(return_value=['fake_output', '']))
        expected = ['docker', 'stop', 'manila-fake-conainer']

        self.DockerExecHelper.stop_container("manila-fake-conainer")

        self.DockerExecHelper._inner_execute.assert_called_once_with(expected)

    def test_stop_container_oh_noes(self):
        self.mock_object(self.DockerExecHelper, "_inner_execute",
                         mock.Mock(side_effect=OSError))

        self.assertRaises(exception.ShareBackendException,
                          self.DockerExecHelper.stop_container,
                          "manila-fake-container")

    def test_execute(self):
        self.mock_object(self.DockerExecHelper, "_inner_execute",
                         mock.Mock(return_value='fake_output'))
        expected = ['docker', 'exec', '-i', 'fake_container', 'fake_script']

        self.DockerExecHelper.execute("fake_container", ["fake_script"])

        self.DockerExecHelper._inner_execute.assert_called_once_with(
            expected, ignore_errors=False)

    def test_execute_name_not_there(self):
        self.assertRaises(exception.ManilaException,
                          self.DockerExecHelper.execute,
                          None, ['do', 'stuff'])

    def test_execute_command_not_there(self):
        self.assertRaises(exception.ManilaException,
                          self.DockerExecHelper.execute,
                          'fake-name', None)

    def test_execute_bad_command_format(self):
        self.assertRaises(exception.ManilaException,
                          self.DockerExecHelper.execute,
                          'fake-name', 'do stuff')

    def test__inner_execute_ok(self):
        self.DockerExecHelper._execute = mock.Mock(return_value='fake')

        result = self.DockerExecHelper._inner_execute("fake_command")

        self.assertEqual(result, 'fake')

    def test__inner_execute_not_ok(self):
        self.DockerExecHelper._execute = mock.Mock(side_effect=[OSError()])

        self.assertRaises(OSError,
                          self.DockerExecHelper._inner_execute, "fake_command")

    def test__inner_execute_not_ok_ignore_errors(self):
        self.DockerExecHelper._execute = mock.Mock(side_effect=OSError())

        result = self.DockerExecHelper._inner_execute("fake_command",
                                                      ignore_errors=True)

        self.assertIsNone(result)

    @ddt.data(('inet',
               "192.168.0.254",
               ["5: br0 inet 192.168.0.254/24 brd 192.168.0.255 "
                "scope global br0 valid_lft forever preferred_lft forever"]),
              ("inet6",
               "2001:470:8:c82:6600:6aff:fe84:8dda",
               ["5: br0 inet6 2001:470:8:c82:6600:6aff:fe84:8dda/64 "
                "scope global valid_lft forever preferred_lft forever"]),
              )
    @ddt.unpack
    def test_fetch_container_address(self, address_family, expected_address,
                                     return_value):
        fake_name = "fakeserver"
        mock_execute = self.DockerExecHelper.execute = mock.Mock(
            return_value=return_value)

        address = self.DockerExecHelper.fetch_container_address(
            fake_name,
            address_family)

        self.assertEqual(expected_address, address)
        mock_execute.assert_called_once_with(
            fake_name, ["ip", "-oneline", "-family", address_family, "address",
                        "show", "scope", "global", "dev", "eth0"]
        )

    def test_rename_container(self):

        fake_old_name = "old_name"
        fake_new_name = "new_name"
        fake_veth_name = "veth_fake"
        self.DockerExecHelper.find_container_veth = mock.Mock(
            return_value=fake_veth_name)
        mock__inner_execute = self.DockerExecHelper._inner_execute = mock.Mock(
            return_value=['fake', ''])

        self.DockerExecHelper.rename_container(fake_old_name, fake_new_name)
        self.DockerExecHelper.find_container_veth.assert_called_once_with(
            fake_old_name
        )
        mock__inner_execute.assert_has_calls([
            mock.call(["docker", "rename", fake_old_name, fake_new_name]),
            mock.call(["ovs-vsctl", "set", "interface", fake_veth_name,
                      "external-ids:manila-container=%s" % fake_new_name])
        ])

    def test_rename_container_exception_veth(self):

        self.DockerExecHelper.find_container_veth = mock.Mock(
            return_value=None)

        self.assertRaises(exception.ManilaException,
                          self.DockerExecHelper.rename_container,
                          "old_name", "new_name")

    @ddt.data([['fake', ''], OSError, ['fake', '']],
              [['fake', ''], OSError, OSError],
              [OSError])
    def test_rename_container_exception_cmds(self, side_effect):
        fake_old_name = "old_name"
        fake_new_name = "new_name"
        fake_veth_name = "veth_fake"

        self.DockerExecHelper.find_container_veth = mock.Mock(
            return_value=fake_veth_name)
        mock__inner_execute = self.DockerExecHelper._inner_execute = mock.Mock(
            side_effect=side_effect)

        self.assertRaises(exception.ShareBackendException,
                          self.DockerExecHelper.rename_container,
                          fake_old_name, fake_new_name)

        if len(side_effect) > 1:
            mock__inner_execute.assert_has_calls([
                mock.call(["docker", "rename", fake_old_name, fake_new_name]),
                mock.call(["ovs-vsctl", "set", "interface", fake_veth_name,
                           "external-ids:manila-container=%s" % fake_new_name])
            ])
        else:
            mock__inner_execute.assert_has_calls([
                mock.call(["docker", "rename", fake_old_name, fake_new_name]),
            ])

    @ddt.data('my_container', 'manila_my_container')
    def test_find_container_veth(self, name):

        interfaces = [fakes.FAKE_VSCTL_LIST_INTERFACE_1,
                      fakes.FAKE_VSCTL_LIST_INTERFACE_2,
                      fakes.FAKE_VSCTL_LIST_INTERFACE_4]

        if 'manila_' in name:
            list_interfaces = [fakes.FAKE_VSCTL_LIST_INTERFACES]
            interfaces.append(fakes.FAKE_VSCTL_LIST_INTERFACE_3)
        else:
            list_interfaces = [fakes.FAKE_VSCTL_LIST_INTERFACES_X]
            interfaces.append(fakes.FAKE_VSCTL_LIST_INTERFACE_3_X)

        def get_interface_data_according_to_veth(*args, **kwargs):
            if len(args) == 4:
                for interface in interfaces:
                    if args[3] in interface:
                        return [interface]
            else:
                return list_interfaces

        self.DockerExecHelper._execute = mock.Mock(
            side_effect=get_interface_data_according_to_veth)

        result = self.DockerExecHelper.find_container_veth(name)

        self.assertEqual("veth3jd83j7", result)

    @ddt.data(True, False)
    def test_find_container_veth_not_found(self, remove_veth):

        if remove_veth:
            list_executes = [[fakes.FAKE_VSCTL_LIST_INTERFACES],
                             [fakes.FAKE_VSCTL_LIST_INTERFACE_1],
                             OSError,
                             [fakes.FAKE_VSCTL_LIST_INTERFACE_3],
                             [fakes.FAKE_VSCTL_LIST_INTERFACE_4]]
        else:
            list_executes = [[fakes.FAKE_VSCTL_LIST_INTERFACES],
                             [fakes.FAKE_VSCTL_LIST_INTERFACE_1],
                             [fakes.FAKE_VSCTL_LIST_INTERFACE_2],
                             [fakes.FAKE_VSCTL_LIST_INTERFACE_3],
                             [fakes.FAKE_VSCTL_LIST_INTERFACE_4]]

        self.DockerExecHelper._execute = mock.Mock(
            side_effect=list_executes)
        list_veths = ['veth11b2c34', 'veth25f6g7h', 'veth3jd83j7',
                      'veth4i9j10k']

        self.assertIsNone(
            self.DockerExecHelper.find_container_veth("foo_bar"))
        list_calls = [mock.call("ovs-vsctl", "list", "interface",
                                run_as_root=True)]

        for veth in list_veths:
            list_calls.append(
                mock.call("ovs-vsctl", "list", "interface", veth,
                          run_as_root=True)
            )

        self.DockerExecHelper._execute.assert_has_calls(
            list_calls, any_order=True
        )

    @ddt.data((["wrong_name\nfake\nfake_container\nfake_name'"], True),
              (["wrong_name\nfake_container\nfake'"], False),
              ("\n", False))
    @ddt.unpack
    def test_container_exists(self, fake_return_value, expected_result):

        self.DockerExecHelper._execute = mock.Mock(
            return_value=fake_return_value)

        result = self.DockerExecHelper.container_exists("fake_name")

        self.DockerExecHelper._execute.assert_called_once_with(
            "docker", "ps", "--no-trunc", "--format='{{.Names}}'",
            run_as_root=True)
        self.assertEqual(expected_result, result)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add 'availability_zone_id' field to 'share_groups' table.

Revision ID: 5237b6625330
Revises: 7d142971c4ef
Create Date: 2017-03-17 18:49:53.742325

"""

# revision identifiers, used by Alembic.
revision = '5237b6625330'
down_revision = '7d142971c4ef'

from alembic import op
import sqlalchemy as sa


SG_TABLE_NAME = 'share_groups'
ATTR_NAME = 'availability_zone_id'


def upgrade():
    op.add_column(
        SG_TABLE_NAME,
        sa.Column(
            ATTR_NAME,
            sa.String(36),
            default=None,
            nullable=True,
        ),
    )


def downgrade():
    op.drop_column(SG_TABLE_NAME, ATTR_NAME)



# Copyright (c) 2015 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""GlusterFS directory mapped share layout."""

import math
import os

from oslo_config import cfg
from oslo_log import log
import six
import xml.etree.cElementTree as etree

from manila import exception
from manila.i18n import _
from manila.share.drivers.glusterfs import common
from manila.share.drivers.glusterfs import layout
from manila import utils

LOG = log.getLogger(__name__)

glusterfs_directory_mapped_opts = [
    cfg.StrOpt('glusterfs_target',
               help='Specifies the GlusterFS volume to be mounted on the '
                    'Manila host. It is of the form '
                    '[remoteuser@]<volserver>:<volid>.'),
    cfg.StrOpt('glusterfs_mount_point_base',
               default='$state_path/mnt',
               help='Base directory containing mount points for Gluster '
                    'volumes.'),
]

CONF = cfg.CONF
CONF.register_opts(glusterfs_directory_mapped_opts)


class GlusterfsDirectoryMappedLayout(layout.GlusterfsShareLayoutBase):

    def __init__(self, driver, *args, **kwargs):
        super(GlusterfsDirectoryMappedLayout, self).__init__(
            driver, *args, **kwargs)
        self.configuration.append_config_values(
            common.glusterfs_common_opts)
        self.configuration.append_config_values(
            glusterfs_directory_mapped_opts)

    def _glustermanager(self, gluster_address):
        """Create GlusterManager object for gluster_address."""

        return common.GlusterManager(
            gluster_address, self.driver._execute,
            self.configuration.glusterfs_path_to_private_key,
            self.configuration.glusterfs_server_password,
            requires={'volume': True})

    def do_setup(self, context):
        """Prepares the backend and appropriate NAS helpers."""
        if not self.configuration.glusterfs_target:
            raise exception.GlusterfsException(
                _('glusterfs_target configuration that specifies the GlusterFS'
                  ' volume to be mounted on the Manila host is not set.'))
        self.gluster_manager = self._glustermanager(
            self.configuration.glusterfs_target)
        self.gluster_manager.check_gluster_version(
            self.driver.GLUSTERFS_VERSION_MIN)
        self._check_mount_glusterfs()

        # enable quota options of a GlusteFS volume to allow
        # creation of shares of specific size
        args = ('volume', 'quota', self.gluster_manager.volume, 'enable')
        try:
            self.gluster_manager.gluster_call(*args)
        except exception.GlusterfsException:
            if (self.gluster_manager.
                    get_vol_option('features.quota')) != 'on':
                LOG.exception("Error in tuning GlusterFS volume to enable "
                              "creation of shares of specific size.")
                raise

        self._ensure_gluster_vol_mounted()

    def _share_manager(self, share):
        comp_path = self.gluster_manager.components.copy()
        comp_path.update({'path': '/' + share['name']})
        return self._glustermanager(comp_path)

    def _get_mount_point_for_gluster_vol(self):
        """Return mount point for the GlusterFS volume."""
        return os.path.join(self.configuration.glusterfs_mount_point_base,
                            self.gluster_manager.volume)

    def _ensure_gluster_vol_mounted(self):
        """Ensure GlusterFS volume is native-mounted on Manila host."""
        mount_path = self._get_mount_point_for_gluster_vol()
        try:
            common._mount_gluster_vol(self.driver._execute,
                                      self.gluster_manager.export, mount_path,
                                      ensure=True)
        except exception.GlusterfsException:
            LOG.exception('Could not mount the Gluster volume %s',
                          self.gluster_manager.volume)
            raise

    def _get_local_share_path(self, share):
        """Determine mount path of the GlusterFS volume in the Manila host."""
        local_vol_path = self._get_mount_point_for_gluster_vol()
        if not os.access(local_vol_path, os.R_OK):
            raise exception.GlusterfsException('share path %s does not exist' %
                                               local_vol_path)
        return os.path.join(local_vol_path, share['name'])

    def _update_share_stats(self):
        """Retrieve stats info from the GlusterFS volume."""

        # sanity check for gluster ctl mount
        smpb = os.stat(self.configuration.glusterfs_mount_point_base)
        smp = os.stat(self._get_mount_point_for_gluster_vol())
        if smpb.st_dev == smp.st_dev:
            raise exception.GlusterfsException(
                _("GlusterFS control mount is not available")
            )
        smpv = os.statvfs(self._get_mount_point_for_gluster_vol())

        return {'total_capacity_gb': (smpv.f_blocks * smpv.f_frsize) >> 30,
                'free_capacity_gb': (smpv.f_bavail * smpv.f_frsize) >> 30}

    def create_share(self, ctx, share, share_server=None):
        """Create a sub-directory/share in the GlusterFS volume."""
        # probe into getting a NAS protocol helper for the share in order
        # to facilitate early detection of unsupported protocol type
        local_share_path = self._get_local_share_path(share)
        cmd = ['mkdir', local_share_path]

        try:
            self.driver._execute(*cmd, run_as_root=True)
            self._set_directory_quota(share, share['size'])
        except Exception as exc:
            if isinstance(exc, exception.ProcessExecutionError):
                exc = exception.GlusterfsException(exc)
            if isinstance(exc, exception.GlusterfsException):
                self._cleanup_create_share(local_share_path, share['name'])
                LOG.error('Unable to create share %s', share['name'])
            raise exc

        comp_share = self.gluster_manager.components.copy()
        comp_share['path'] = '/' + share['name']
        export_location = self.driver._setup_via_manager(
            {'share': share,
             'manager': self._glustermanager(comp_share)})

        return export_location

    def _cleanup_create_share(self, share_path, share_name):
        """Cleanup share that errored out during its creation."""
        if os.path.exists(share_path):
            cmd = ['rm', '-rf', share_path]
            try:
                self.driver._execute(*cmd, run_as_root=True)
            except exception.ProcessExecutionError as exc:
                LOG.error('Cannot cleanup share, %s, that errored out '
                          'during its creation, but exists in GlusterFS '
                          'volume.', share_name)
                raise exception.GlusterfsException(exc)

    def delete_share(self, context, share, share_server=None):
        """Remove a sub-directory/share from the GlusterFS volume."""
        local_share_path = self._get_local_share_path(share)
        cmd = ['rm', '-rf', local_share_path]
        try:
            self.driver._execute(*cmd, run_as_root=True)
        except exception.ProcessExecutionError:
            LOG.exception('Unable to delete share %s', share['name'])
            raise

    def ensure_share(self, context, share, share_server=None):
        pass

    def create_share_from_snapshot(self, context, share, snapshot,
                                   share_server=None, parent_share=None):
        raise NotImplementedError

    def create_snapshot(self, context, snapshot, share_server=None):
        raise NotImplementedError

    def delete_snapshot(self, context, snapshot, share_server=None):
        raise NotImplementedError

    def manage_existing(self, share, driver_options):
        raise NotImplementedError

    def unmanage(self, share):
        raise NotImplementedError

    def extend_share(self, share, new_size, share_server=None):
        """Extend a sub-directory/share in the GlusterFS volume."""
        self._set_directory_quota(share, new_size)

    def shrink_share(self, share, new_size, share_server=None):
        """Shrink a sub-directory/share in the GlusterFS volume."""
        usage = self._get_directory_usage(share)
        consumed_limit = int(math.ceil(usage))
        if consumed_limit > new_size:
            raise exception.ShareShrinkingPossibleDataLoss(
                share_id=share['id'])

        self._set_directory_quota(share, new_size)

    def _set_directory_quota(self, share, new_size):
        sizestr = six.text_type(new_size) + 'GB'
        share_dir = '/' + share['name']

        args = ('volume', 'quota', self.gluster_manager.volume,
                'limit-usage', share_dir, sizestr)

        try:
            self.gluster_manager.gluster_call(*args)
        except exception.GlusterfsException:
            LOG.error('Unable to set quota share %s', share['name'])
            raise

    def _get_directory_usage(self, share):
        share_dir = '/' + share['name']

        args = ('--xml', 'volume', 'quota', self.gluster_manager.volume,
                'list', share_dir)

        try:
            out, err = self.gluster_manager.gluster_call(*args)
        except exception.GlusterfsException:
            LOG.error('Unable to get quota share %s', share['name'])
            raise

        volxml = etree.fromstring(out)
        usage_byte = volxml.find('./volQuota/limit/used_space').text
        usage = utils.translate_string_size_to_float(usage_byte)

        return usage



# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import alembic

from manila.db import migration
from manila import test


class MigrationTestCase(test.TestCase):
    def setUp(self):
        super(MigrationTestCase, self).setUp()
        self.config_patcher = mock.patch(
            'manila.db.migrations.alembic.migration._alembic_config')
        self.config = self.config_patcher.start()
        self.config.return_value = 'fake_config'
        self.addCleanup(self.config_patcher.stop)

    @mock.patch('alembic.command.upgrade')
    def test_upgrade(self, upgrade):
        migration.upgrade('version_1')
        upgrade.assert_called_once_with('fake_config', 'version_1')

    @mock.patch('alembic.command.upgrade')
    def test_upgrade_none_version(self, upgrade):
        migration.upgrade(None)
        upgrade.assert_called_once_with('fake_config', 'head')

    @mock.patch('alembic.command.downgrade')
    def test_downgrade(self, downgrade):
        migration.downgrade('version_1')
        downgrade.assert_called_once_with('fake_config', 'version_1')

    @mock.patch('alembic.command.downgrade')
    def test_downgrade_none_version(self, downgrade):
        migration.downgrade(None)
        downgrade.assert_called_once_with('fake_config', 'base')

    @mock.patch('alembic.command.stamp')
    def test_stamp(self, stamp):
        migration.stamp('version_1')
        stamp.assert_called_once_with('fake_config', 'version_1')

    @mock.patch('alembic.command.stamp')
    def test_stamp_none_version(self, stamp):
        migration.stamp(None)
        stamp.assert_called_once_with('fake_config', 'head')

    @mock.patch('alembic.command.revision')
    def test_revision(self, revision):
        migration.revision('test_message', 'autogenerate_value')
        revision.assert_called_once_with('fake_config', 'test_message',
                                         'autogenerate_value')

    @mock.patch.object(alembic.migration.MigrationContext, 'configure',
                       mock.Mock())
    def test_version(self):
        context = mock.Mock()
        context.get_current_revision = mock.Mock()
        alembic.migration.MigrationContext.configure.return_value = context
        migration.version()
        context.get_current_revision.assert_called_once_with()



# Copyright 2015 Intel, Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Unit tests for HDFS native protocol driver module."""

import socket
from unittest import mock

from oslo_concurrency import processutils
from oslo_config import cfg
import six

from manila import context
from manila import exception
import manila.share.configuration as config
import manila.share.drivers.hdfs.hdfs_native as hdfs_native
from manila import test
from manila.tests import fake_share
from manila import utils


CONF = cfg.CONF


class HDFSNativeShareDriverTestCase(test.TestCase):
    """Tests HDFSNativeShareDriver."""

    def setUp(self):
        super(HDFSNativeShareDriverTestCase, self).setUp()
        self._context = context.get_admin_context()
        self._hdfs_execute = mock.Mock(return_value=('', ''))
        self.local_ip = '192.168.1.1'

        CONF.set_default('driver_handles_share_servers', False)
        CONF.set_default('hdfs_namenode_ip', self.local_ip)
        CONF.set_default('hdfs_ssh_name', 'fake_sshname')
        CONF.set_default('hdfs_ssh_pw', 'fake_sshpw')
        CONF.set_default('hdfs_ssh_private_key', 'fake_sshkey')

        self.fake_conf = config.Configuration(None)
        self._driver = hdfs_native.HDFSNativeShareDriver(
            execute=self._hdfs_execute,
            configuration=self.fake_conf)
        self.hdfs_bin = 'hdfs'
        self._driver._hdfs_bin = 'fake_hdfs_bin'
        self.share = fake_share.fake_share(share_proto='HDFS')
        self.snapshot = fake_share.fake_snapshot(share_proto='HDFS')
        self.access = fake_share.fake_access(access_type='user')
        self.fakesharepath = 'hdfs://1.2.3.4:5/share-0'
        self.fakesnapshotpath = '/share-0/.snapshot/snapshot-0'

        socket.gethostname = mock.Mock(return_value='testserver')
        socket.gethostbyname_ex = mock.Mock(return_value=(
            'localhost',
            ['localhost.localdomain', 'testserver'],
            ['127.0.0.1', self.local_ip]))

    def test_do_setup(self):
        self._driver.do_setup(self._context)
        self.assertEqual(self._driver._hdfs_bin, self.hdfs_bin)

    def test_create_share(self):
        self._driver._create_share = mock.Mock()
        self._driver._get_share_path = mock.Mock(
            return_value=self.fakesharepath)
        result = self._driver.create_share(self._context, self.share,
                                           share_server=None)
        self._driver._create_share.assert_called_once_with(self.share)
        self._driver._get_share_path.assert_called_once_with(self.share)

        self.assertEqual(self.fakesharepath, result)

    def test_create_share_unsupported_proto(self):
        self._driver._get_share_path = mock.Mock()
        self.assertRaises(exception.HDFSException,
                          self._driver.create_share,
                          self._context,
                          fake_share.fake_share(),
                          share_server=None)
        self.assertFalse(self._driver._get_share_path.called)

    def test__set_share_size(self):
        share_dir = '/' + self.share['name']
        sizestr = six.text_type(self.share['size']) + 'g'
        self._driver._hdfs_execute = mock.Mock(return_value=True)
        self._driver._set_share_size(self.share)
        self._driver._hdfs_execute.assert_called_once_with(
            'fake_hdfs_bin', 'dfsadmin', '-setSpaceQuota', sizestr, share_dir)

    def test__set_share_size_exception(self):
        share_dir = '/' + self.share['name']
        sizestr = six.text_type(self.share['size']) + 'g'
        self._driver._hdfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError)
        self.assertRaises(exception.HDFSException,
                          self._driver._set_share_size, self.share)
        self._driver._hdfs_execute.assert_called_once_with(
            'fake_hdfs_bin', 'dfsadmin', '-setSpaceQuota', sizestr, share_dir)

    def test__set_share_size_with_new_size(self):
        share_dir = '/' + self.share['name']
        new_size = 'fake_size'
        sizestr = new_size + 'g'
        self._driver._hdfs_execute = mock.Mock(return_value=True)
        self._driver._set_share_size(self.share, new_size)
        self._driver._hdfs_execute.assert_called_once_with(
            'fake_hdfs_bin', 'dfsadmin', '-setSpaceQuota', sizestr, share_dir)

    def test__create_share(self):
        share_dir = '/' + self.share['name']
        self._driver._hdfs_execute = mock.Mock(return_value=True)
        self._driver._set_share_size = mock.Mock()
        self._driver._create_share(self.share)
        self._driver._hdfs_execute.assert_any_call(
            'fake_hdfs_bin', 'dfs', '-mkdir', share_dir)
        self._driver._set_share_size.assert_called_once_with(self.share)
        self._driver._hdfs_execute.assert_any_call(
            'fake_hdfs_bin', 'dfsadmin', '-allowSnapshot', share_dir)

    def test__create_share_exception(self):
        share_dir = '/' + self.share['name']
        self._driver._hdfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError)
        self.assertRaises(exception.HDFSException,
                          self._driver._create_share, self.share)
        self._driver._hdfs_execute.assert_called_once_with(
            'fake_hdfs_bin', 'dfs', '-mkdir', share_dir)

    def test_create_share_from_empty_snapshot(self):
        return_hdfs_execute = (None, None)
        self._driver._hdfs_execute = mock.Mock(
            return_value=return_hdfs_execute)
        self._driver._create_share = mock.Mock(return_value=True)
        self._driver._get_share_path = mock.Mock(return_value=self.
                                                 fakesharepath)
        self._driver._get_snapshot_path = mock.Mock(return_value=self.
                                                    fakesnapshotpath)
        result = self._driver.create_share_from_snapshot(self._context,
                                                         self.share,
                                                         self.snapshot,
                                                         share_server=None)
        self._driver._create_share.assert_called_once_with(self.share)
        self._driver._get_snapshot_path.assert_called_once_with(
            self.snapshot)
        self._driver._hdfs_execute.assert_called_once_with(
            'fake_hdfs_bin', 'dfs', '-ls', self.fakesnapshotpath)
        self._driver._get_share_path.assert_called_once_with(self.share)
        self.assertEqual(self.fakesharepath, result)

    def test_create_share_from_snapshot(self):
        return_hdfs_execute = ("fake_content", None)
        self._driver._hdfs_execute = mock.Mock(
            return_value=return_hdfs_execute)
        self._driver._create_share = mock.Mock(return_value=True)
        self._driver._get_share_path = mock.Mock(return_value=self.
                                                 fakesharepath)
        self._driver._get_snapshot_path = mock.Mock(return_value=self.
                                                    fakesnapshotpath)
        result = self._driver.create_share_from_snapshot(self._context,
                                                         self.share,
                                                         self.snapshot,
                                                         share_server=None)
        self._driver._create_share.assert_called_once_with(self.share)
        self._driver._get_snapshot_path.assert_called_once_with(
            self.snapshot)

        calls = [mock.call('fake_hdfs_bin', 'dfs',
                           '-ls', self.fakesnapshotpath),
                 mock.call('fake_hdfs_bin', 'dfs', '-cp',
                           self.fakesnapshotpath + '/*',
                           '/' + self.share['name'])]

        self._driver._hdfs_execute.assert_has_calls(calls)
        self._driver._get_share_path.assert_called_once_with(self.share)
        self.assertEqual(self.fakesharepath, result)

    def test_create_share_from_snapshot_exception(self):
        self._driver._create_share = mock.Mock(return_value=True)
        self._driver._get_snapshot_path = mock.Mock(return_value=self.
                                                    fakesnapshotpath)
        self._driver._get_share_path = mock.Mock(return_value=self.
                                                 fakesharepath)
        self._driver._hdfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError)
        self.assertRaises(exception.HDFSException,
                          self._driver.create_share_from_snapshot,
                          self._context, self.share,
                          self.snapshot, share_server=None)
        self._driver._create_share.assert_called_once_with(self.share)
        self._driver._get_snapshot_path.assert_called_once_with(self.snapshot)

        self._driver._hdfs_execute.assert_called_once_with(
            'fake_hdfs_bin', 'dfs', '-ls', self.fakesnapshotpath)
        self.assertFalse(self._driver._get_share_path.called)

    def test_create_snapshot(self):
        self._driver._hdfs_execute = mock.Mock(return_value=True)
        self._driver.create_snapshot(self._context, self.snapshot,
                                     share_server=None)
        self._driver._hdfs_execute.assert_called_once_with(
            'fake_hdfs_bin', 'dfs', '-createSnapshot',
            '/' + self.snapshot['share_name'], self.snapshot['name'])

    def test_create_snapshot_exception(self):
        self._driver._hdfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError)
        self.assertRaises(exception.HDFSException,
                          self._driver.create_snapshot, self._context,
                          self.snapshot, share_server=None)
        self._driver._hdfs_execute.assert_called_once_with(
            'fake_hdfs_bin', 'dfs', '-createSnapshot',
            '/' + self.snapshot['share_name'], self.snapshot['name'])

    def test_delete_share(self):
        self._driver._hdfs_execute = mock.Mock(return_value=True)
        self._driver.delete_share(self._context,
                                  self.share,
                                  share_server=None)
        self._driver._hdfs_execute.assert_called_once_with(
            'fake_hdfs_bin', 'dfs', '-rm', '-r',
            '/' + self.share['name'])

    def test_delete_share_exception(self):
        self._driver._hdfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError)
        self.assertRaises(exception.HDFSException,
                          self._driver.delete_share,
                          self._context,
                          self.share,
                          share_server=None)
        self._driver._hdfs_execute.assert_called_once_with(
            'fake_hdfs_bin', 'dfs', '-rm', '-r',
            '/' + self.share['name'])

    def test_delete_snapshot(self):
        self._driver._hdfs_execute = mock.Mock(return_value=True)
        self._driver.delete_snapshot(self._context,
                                     self.snapshot,
                                     share_server=None)
        self._driver._hdfs_execute.assert_called_once_with(
            'fake_hdfs_bin', 'dfs', '-deleteSnapshot',
            '/' + self.snapshot['share_name'], self.snapshot['name'])

    def test_delete_snapshot_exception(self):
        self._driver._hdfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError)
        self.assertRaises(exception.HDFSException,
                          self._driver.delete_snapshot,
                          self._context,
                          self.snapshot,
                          share_server=None)
        self._driver._hdfs_execute.assert_called_once_with(
            'fake_hdfs_bin', 'dfs', '-deleteSnapshot',
            '/' + self.snapshot['share_name'], self.snapshot['name'])

    def test_allow_access(self):
        self._driver._hdfs_execute = mock.Mock(
            return_value=['', ''])
        share_dir = '/' + self.share['name']
        user_access = ':'.join([self.access['access_type'],
                                self.access['access_to'],
                                'rwx'])
        cmd = ['fake_hdfs_bin', 'dfs', '-setfacl', '-m', '-R',
               user_access, share_dir]
        self._driver.allow_access(self._context, self.share, self.access,
                                  share_server=None)
        self._driver._hdfs_execute.assert_called_once_with(
            *cmd, check_exit_code=True)

    def test_allow_access_invalid_access_type(self):
        self.assertRaises(exception.InvalidShareAccess,
                          self._driver.allow_access,
                          self._context,
                          self.share,
                          fake_share.fake_access(
                              access_type='invalid_access_type'),
                          share_server=None)

    def test_allow_access_invalid_access_level(self):
        self.assertRaises(exception.InvalidShareAccess,
                          self._driver.allow_access,
                          self._context,
                          self.share,
                          fake_share.fake_access(
                              access_level='invalid_access_level'),
                          share_server=None)

    def test_allow_access_exception(self):
        self._driver._hdfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError)
        share_dir = '/' + self.share['name']
        user_access = ':'.join([self.access['access_type'],
                                self.access['access_to'],
                                'rwx'])
        cmd = ['fake_hdfs_bin', 'dfs', '-setfacl', '-m', '-R',
               user_access, share_dir]
        self.assertRaises(exception.HDFSException,
                          self._driver.allow_access,
                          self._context,
                          self.share,
                          self.access,
                          share_server=None)
        self._driver._hdfs_execute.assert_called_once_with(
            *cmd, check_exit_code=True)

    def test_deny_access(self):
        self._driver._hdfs_execute = mock.Mock(return_value=['', ''])
        share_dir = '/' + self.share['name']
        access_name = ':'.join([self.access['access_type'],
                                self.access['access_to']])
        cmd = ['fake_hdfs_bin', 'dfs', '-setfacl', '-x', '-R',
               access_name, share_dir]
        self._driver.deny_access(self._context,
                                 self.share,
                                 self.access,
                                 share_server=None)
        self._driver._hdfs_execute.assert_called_once_with(
            *cmd, check_exit_code=True)

    def test_deny_access_exception(self):
        self._driver._hdfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError)
        share_dir = '/' + self.share['name']
        access_name = ':'.join([self.access['access_type'],
                                self.access['access_to']])
        cmd = ['fake_hdfs_bin', 'dfs', '-setfacl', '-x', '-R',
               access_name, share_dir]
        self.assertRaises(exception.HDFSException,
                          self._driver.deny_access,
                          self._context,
                          self.share,
                          self.access,
                          share_server=None)
        self._driver._hdfs_execute.assert_called_once_with(
            *cmd, check_exit_code=True)

    def test_extend_share(self):
        new_size = "fake_size"
        self._driver._set_share_size = mock.Mock()
        self._driver.extend_share(self.share, new_size)
        self._driver._set_share_size.assert_called_once_with(
            self.share, new_size)

    def test__check_hdfs_state_healthy(self):
        fake_out = "fakeinfo\n...Status: HEALTHY"
        self._driver._hdfs_execute = mock.Mock(return_value=(fake_out, ''))
        result = self._driver._check_hdfs_state()
        self._driver._hdfs_execute.assert_called_once_with(
            'fake_hdfs_bin', 'fsck', '/')
        self.assertTrue(result)

    def test__check_hdfs_state_down(self):
        fake_out = "fakeinfo\n...Status: DOWN"
        self._driver._hdfs_execute = mock.Mock(return_value=(fake_out, ''))
        result = self._driver._check_hdfs_state()
        self._driver._hdfs_execute.assert_called_once_with(
            'fake_hdfs_bin', 'fsck', '/')
        self.assertFalse(result)

    def test__check_hdfs_state_exception(self):
        self._driver._hdfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError)
        self.assertRaises(exception.HDFSException,
                          self._driver._check_hdfs_state)
        self._driver._hdfs_execute.assert_called_once_with(
            'fake_hdfs_bin', 'fsck', '/')

    def test__get_available_capacity(self):
        fake_out = ('Configured Capacity: 2.4\n' +
                    'Total Capacity: 2\n' +
                    'DFS free: 1')
        self._driver._hdfs_execute = mock.Mock(return_value=(fake_out, ''))
        total, free = self._driver._get_available_capacity()
        self._driver._hdfs_execute.assert_called_once_with(
            'fake_hdfs_bin', 'dfsadmin', '-report')
        self.assertEqual(2, total)
        self.assertEqual(1, free)

    def test__get_available_capacity_exception(self):
        self._driver._hdfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError)
        self.assertRaises(exception.HDFSException,
                          self._driver._get_available_capacity)
        self._driver._hdfs_execute.assert_called_once_with(
            'fake_hdfs_bin', 'dfsadmin', '-report')

    def test_get_share_stats_refresh_false(self):
        self._driver._stats = {'fake_key': 'fake_value'}
        result = self._driver.get_share_stats(False)
        self.assertEqual(self._driver._stats, result)

    def test_get_share_stats_refresh_true(self):
        self._driver._get_available_capacity = mock.Mock(
            return_value=(11111.0, 12345.0))
        result = self._driver.get_share_stats(True)
        expected_keys = [
            'qos', 'driver_version', 'share_backend_name',
            'free_capacity_gb', 'total_capacity_gb',
            'driver_handles_share_servers',
            'reserved_percentage', 'vendor_name', 'storage_protocol',
            'ipv4_support', 'ipv6_support'
        ]
        for key in expected_keys:
            self.assertIn(key, result)
        self.assertTrue(result['ipv4_support'])
        self.assertFalse(result['ipv6_support'])
        self.assertEqual('HDFS', result['storage_protocol'])
        self._driver._get_available_capacity.assert_called_once_with()

    def test__hdfs_local_execute(self):
        cmd = 'testcmd'
        self.mock_object(utils, 'execute', mock.Mock(return_value=True))
        self._driver._hdfs_local_execute(cmd)
        utils.execute.assert_called_once_with(cmd, run_as_root=False)

    def test__hdfs_remote_execute(self):
        self._driver._run_ssh = mock.Mock(return_value=True)
        cmd = 'testcmd'
        self._driver._hdfs_remote_execute(cmd, check_exit_code=True)
        self._driver._run_ssh.assert_called_once_with(
            self.local_ip, tuple([cmd]), True)

    def test__run_ssh(self):
        ssh_output = 'fake_ssh_output'
        cmd_list = ['fake', 'cmd']
        ssh = mock.Mock()
        ssh.get_transport = mock.Mock()
        ssh.get_transport().is_active = mock.Mock(return_value=True)
        ssh_pool = mock.Mock()
        ssh_pool.create = mock.Mock(return_value=ssh)
        self.mock_object(utils, 'SSHPool', mock.Mock(return_value=ssh_pool))
        self.mock_object(processutils, 'ssh_execute',
                         mock.Mock(return_value=ssh_output))
        result = self._driver._run_ssh(self.local_ip, cmd_list)
        utils.SSHPool.assert_called_once_with(
            self._driver.configuration.hdfs_namenode_ip,
            self._driver.configuration.hdfs_ssh_port,
            self._driver.configuration.ssh_conn_timeout,
            self._driver.configuration.hdfs_ssh_name,
            password=self._driver.configuration.hdfs_ssh_pw,
            privatekey=self._driver.configuration.hdfs_ssh_private_key,
            min_size=self._driver.configuration.ssh_min_pool_conn,
            max_size=self._driver.configuration.ssh_max_pool_conn)
        ssh_pool.create.assert_called_once_with()
        ssh.get_transport().is_active.assert_called_once_with()
        processutils.ssh_execute.assert_called_once_with(
            ssh, 'fake cmd', check_exit_code=False)
        self.assertEqual(ssh_output, result)

    def test__run_ssh_exception(self):
        cmd_list = ['fake', 'cmd']
        ssh = mock.Mock()
        ssh.get_transport = mock.Mock()
        ssh.get_transport().is_active = mock.Mock(return_value=True)
        ssh_pool = mock.Mock()
        ssh_pool.create = mock.Mock(return_value=ssh)
        self.mock_object(utils, 'SSHPool', mock.Mock(return_value=ssh_pool))
        self.mock_object(processutils, 'ssh_execute',
                         mock.Mock(side_effect=Exception))
        self.assertRaises(exception.HDFSException,
                          self._driver._run_ssh,
                          self.local_ip,
                          cmd_list)
        utils.SSHPool.assert_called_once_with(
            self._driver.configuration.hdfs_namenode_ip,
            self._driver.configuration.hdfs_ssh_port,
            self._driver.configuration.ssh_conn_timeout,
            self._driver.configuration.hdfs_ssh_name,
            password=self._driver.configuration.hdfs_ssh_pw,
            privatekey=self._driver.configuration.hdfs_ssh_private_key,
            min_size=self._driver.configuration.ssh_min_pool_conn,
            max_size=self._driver.configuration.ssh_max_pool_conn)
        ssh_pool.create.assert_called_once_with()
        ssh.get_transport().is_active.assert_called_once_with()
        processutils.ssh_execute.assert_called_once_with(
            ssh, 'fake cmd', check_exit_code=False)



# Copyright 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add required extra spec

Revision ID: 59eb64046740
Revises: 162a3e673105
Create Date: 2015-01-29 15:33:25.348140

"""

# revision identifiers, used by Alembic.
revision = '59eb64046740'
down_revision = '4ee2cf4be19a'

from alembic import op
from oslo_utils import timeutils
import sqlalchemy as sa
from sqlalchemy.sql import table


def upgrade():
    session = sa.orm.Session(bind=op.get_bind().connect())

    es_table = table(
        'share_type_extra_specs',
        sa.Column('created_at', sa.DateTime),
        sa.Column('deleted', sa.Integer),
        sa.Column('share_type_id', sa.String(length=36)),
        sa.Column('spec_key', sa.String(length=255)),
        sa.Column('spec_value', sa.String(length=255)))

    st_table = table(
        'share_types',
        sa.Column('deleted', sa.Integer),
        sa.Column('id', sa.Integer))

    # NOTE(vponomaryov): field 'deleted' is integer here.
    existing_required_extra_specs = (session.query(es_table).
                                     filter(es_table.c.spec_key ==
                                            'driver_handles_share_servers').
                                     filter(es_table.c.deleted == 0).
                                     all())
    exclude_st_ids = [es.share_type_id for es in existing_required_extra_specs]

    # NOTE(vponomaryov): field 'deleted' is string here.
    share_types = (session.query(st_table).
                   filter(st_table.c.deleted.in_(('0', 'False', ))).
                   filter(st_table.c.id.notin_(exclude_st_ids)).
                   all())

    extra_specs = []
    for st in share_types:
        extra_specs.append({
            'spec_key': 'driver_handles_share_servers',
            'spec_value': 'True',
            'deleted': 0,
            'created_at': timeutils.utcnow(),
            'share_type_id': st.id,
        })

    op.bulk_insert(es_table, extra_specs)
    session.close_all()


def downgrade():
    """Downgrade method.

    We can't determine, which extra specs should be removed after insertion,
    that's why do nothing here.
    """



# Copyright 2017 Veritas Technologies LLC.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Unit tests for Veritas Manila driver.
"""
import hashlib
import json
from unittest import mock

from oslo_config import cfg
import requests
import six

from manila import context
from manila import exception
from manila.share import configuration as conf
from manila.share.drivers.veritas import veritas_isa
from manila import test


CONF = cfg.CONF
FAKE_BACKEND = 'fake_backend'


class MockResponse(object):
    def __init__(self):
        self.status_code = 200

    def json(self):
        data = {'fake_key': 'fake_val'}
        return json.dumps(data)


class ACCESSShareDriverTestCase(test.TestCase):
    """Tests ACCESSShareDriver."""

    share = {
        'id': 'fakeid',
        'name': 'fakename',
        'size': 1,
        'share_proto': 'NFS',
        'export_locations': [{'path': '10.20.30.40:/vx/fake_location'}],
        'snapshot_id': False
    }

    share2 = {
        'id': 'fakeid2',
        'name': 'fakename2',
        'size': 4,
        'share_proto': 'NFS',
    }

    share3 = {
        'id': 'fakeid3',
        'name': 'fakename3',
        'size': 2,
        'share_proto': 'NFS',
        'export_location': '/vx/fake_location',
        'snapshot_id': True
    }

    snapshot = {
        'id': 'fakesnapshotid',
        'share_name': 'fakename',
        'share_id': 'fakeid',
        'name': 'fakesnapshotname',
        'share_size': 1,
        'share_proto': 'NFS',
        'snapshot_id': 'fake_snap_id',
    }

    access = {
        'id': 'fakeaccid',
        'access_type': 'ip',
        'access_to': '10.0.0.2',
        'access_level': 'rw',
        'state': 'active',
    }

    access2 = {
        'id': 'fakeaccid2',
        'access_type': 'user',
        'access_to': '10.0.0.3',
        'access_level': 'rw',
        'state': 'active',
    }

    access3 = {
        'id': 'fakeaccid3',
        'access_type': 'ip',
        'access_to': '10.0.0.4',
        'access_level': 'rw+',
        'state': 'active',
    }

    access4 = {
        'id': 'fakeaccid',
        'access_type': 'ip',
        'access_to': '10.0.0.2',
        'access_level': 'ro',
        'state': 'active',
    }

    def setUp(self):
        super(ACCESSShareDriverTestCase, self).setUp()
        self._create_fake_config()
        lcfg = self.configuration
        self._context = context.get_admin_context()
        self._driver = veritas_isa.ACCESSShareDriver(False, configuration=lcfg)
        self._driver.do_setup(self._context)

    def _create_fake_config(self):
        def _safe_get(opt):
            return getattr(self.configuration, opt)

        self.mock_object(veritas_isa.ACCESSShareDriver, '_authenticate_access')
        self.configuration = mock.Mock(spec=conf.Configuration)
        self.configuration.safe_get = mock.Mock(side_effect=_safe_get)
        self.configuration.va_server_ip = '1.1.1.1'
        self.configuration.va_pool = 'pool1'
        self.configuration.va_user = 'user'
        self.configuration.va_pwd = 'passwd'
        self.configuration.va_port = 14161
        self.configuration.va_ssl = 'False'
        self.configuration.va_fstype = 'simple'
        self.configuration.network_config_group = 'fake_network_config_group'
        self.configuration.admin_network_config_group = (
            'fake_admin_network_config_group')
        self.configuration.driver_handles_share_servers = False
        self.configuration.share_backend_name = FAKE_BACKEND
        self.configuration.replication_domain = 'Disable'
        self.configuration.filter_function = 'Disable'
        self.configuration.goodness_function = 'Disable'
        self.configuration.max_shares_per_share_server = -1
        self.configuration.max_share_server_size = -1

    def test_create_share(self):
        self.mock_object(self._driver, '_get_va_share_name')
        self.mock_object(self._driver, '_get_va_share_path')
        self.mock_object(self._driver, '_get_vip')
        self.mock_object(self._driver, '_access_api')
        length = len(self.share['name'])
        index = int(length / 2)
        name1 = self.share['name'][:index]
        name2 = self.share['name'][index:]
        crc1 = hashlib.md5(name1.encode('utf-8')).hexdigest()[:8]
        crc2 = hashlib.md5(name2.encode('utf-8')).hexdigest()[:8]

        share_name_to_ret = crc1 + '-' + crc2
        share_path_to_ret = '/vx/' + crc1 + '-' + crc2

        self._driver._get_va_share_name.return_value = share_name_to_ret
        self._driver._get_va_share_path.return_value = share_path_to_ret

        self._driver._get_vip.return_value = '1.1.1.1'
        self._driver.create_share(self._context, self.share)

        self.assertEqual(1, self._driver._get_vip.call_count)
        self.assertEqual(1, self._driver._get_va_share_name.call_count)
        self.assertEqual(1, self._driver._get_va_share_path.call_count)

    def test_create_share_negative(self):
        self.mock_object(self._driver, '_access_api')

        self._driver._access_api.return_value = False

        self.assertRaises(exception.ShareBackendException,
                          self._driver.create_share,
                          self._context,
                          self.share)

    def test_create_share_from_snapshot(self):
        self.mock_object(self._driver, '_get_vip')

        sharename = self._driver._get_va_share_name(
            self.snapshot['share_name'])
        snapname = self._driver._get_va_snap_name(self.snapshot['name'])
        sharepath = self._driver._get_va_share_path(sharename)
        self._driver._get_vip.return_value = '1.1.1.1'
        vip = self._driver._get_vip()
        location = (six.text_type(vip) + ':' +
                    six.text_type(sharepath) + ':' + six.text_type(snapname))

        ret = self._driver.create_share_from_snapshot(self._context,
                                                      self.share,
                                                      self.snapshot)
        self.assertEqual(location, ret)

    def test_delete_share(self):
        self.mock_object(self._driver, '_access_api')
        self.mock_object(self._driver, '_does_item_exist_at_va_backend')
        self._driver._does_item_exist_at_va_backend.return_value = True
        self._driver.delete_share(self._context, self.share)
        self.assertEqual(2, self._driver._access_api.call_count)

    def test_delete_share_negative(self):
        self.mock_object(self._driver, '_access_api')
        self.mock_object(self._driver, '_does_item_exist_at_va_backend')

        self._driver._does_item_exist_at_va_backend.return_value = True
        self._driver._access_api.return_value = False

        self.assertRaises(exception.ShareBackendException,
                          self._driver.delete_share,
                          self._context, self.share)

    def test_delete_share_if_share_created_from_snap(self):
        self.mock_object(self._driver, '_access_api')
        self.mock_object(self._driver, '_does_item_exist_at_va_backend')

        self._driver.delete_share(self._context, self.share3)
        self.assertEqual(0,
                         (self._driver.
                          _does_item_exist_at_va_backend.call_count))
        self.assertEqual(0, self._driver._access_api.call_count)

    def test_delete_share_if_not_present_at_backend(self):
        self.mock_object(self._driver, '_does_item_exist_at_va_backend')
        self.mock_object(self._driver, '_access_api')

        self._driver._does_item_exist_at_va_backend.return_value = False
        self._driver.delete_share(self._context, self.share)
        self.assertEqual(1,
                         (self._driver.
                          _does_item_exist_at_va_backend.call_count))
        self.assertEqual(0, self._driver._access_api.call_count)

    def test_create_snapshot(self):
        self.mock_object(self._driver, '_access_api')
        self._driver.create_snapshot(self._context, self.snapshot)
        self.assertEqual(2, self._driver._access_api.call_count)

    def test_create_snapshot_negative(self):
        self.mock_object(self._driver, '_access_api')

        self._driver._access_api.return_value = False

        self.assertRaises(exception.ShareBackendException,
                          self._driver.create_snapshot,
                          self._context,
                          self.snapshot)

    def test_delete_snapshot(self):
        self.mock_object(self._driver, '_access_api')
        self.mock_object(self._driver, '_does_item_exist_at_va_backend')

        self._driver._does_item_exist_at_va_backend.return_value = True
        self._driver.delete_snapshot(self._context, self.snapshot)
        self.assertEqual(2, self._driver._access_api.call_count)

    def test_delete_snapshot_negative(self):
        self.mock_object(self._driver, '_access_api')
        self.mock_object(self._driver, '_does_item_exist_at_va_backend')

        self._driver._does_item_exist_at_va_backend.return_value = True
        self._driver._access_api.return_value = False

        self.assertRaises(exception.ShareBackendException,
                          self._driver.delete_snapshot,
                          self._context, self.snapshot)

    def test_delete_snapshot_if_not_present_at_backend(self):
        self.mock_object(self._driver, '_does_item_exist_at_va_backend')
        self.mock_object(self._driver, '_access_api')

        self._driver._does_item_exist_at_va_backend.return_value = False
        self._driver.delete_snapshot(self._context, self.snapshot)
        self.assertEqual(1,
                         (self._driver.
                          _does_item_exist_at_va_backend.call_count))
        self.assertEqual(0, self._driver._access_api.call_count)

    def test_update_access_for_allow(self):
        self.mock_object(self._driver, '_access_api')
        self._driver.update_access(self._context, self.share, [],
                                   [self.access], [])
        self.assertEqual(2, self._driver._access_api.call_count)

    def test_update_access_for_allow_negative(self):
        self.mock_object(self._driver, '_access_api')
        self._driver._access_api.return_value = False
        self.assertRaises(exception.ShareBackendException,
                          self._driver.update_access,
                          self._context,
                          self.share, [], [self.access], [])

        self.assertRaises(exception.InvalidShareAccess,
                          self._driver.update_access,
                          self._context,
                          self.share, [], [self.access2], [])

        self.assertRaises(exception.InvalidShareAccessLevel,
                          self._driver.update_access,
                          self._context,
                          self.share, [], [self.access3], [])

    def test_update_access_for_deny(self):
        self.mock_object(self._driver, '_access_api')
        self._driver.update_access(self._context, self.share,
                                   [], [], [self.access])
        self.assertEqual(2, self._driver._access_api.call_count)

    def test_update_access_for_deny_negative(self):
        self.mock_object(self._driver, '_access_api')
        self._driver._access_api.return_value = False
        self.assertRaises(exception.ShareBackendException,
                          self._driver.update_access,
                          self._context,
                          self.share, [], [], [self.access])

    def test_update_access_for_deny_for_invalid_access_type(self):
        self.mock_object(self._driver, '_access_api')
        self._driver.update_access(self._context, self.share,
                                   [], [], [self.access2])
        self.assertEqual(0, self._driver._access_api.call_count)

    def test_update_access_for_empty_rule_list(self):
        self.mock_object(self._driver, '_allow_access')
        self.mock_object(self._driver, '_deny_access')
        self._driver.update_access(self._context, self.share,
                                   [], [], [])
        self.assertEqual(0, self._driver._allow_access.call_count)
        self.assertEqual(0, self._driver._deny_access.call_count)

    def test_update_access_for_access_rules(self):
        self.mock_object(self._driver, '_fetch_existing_rule')
        self.mock_object(self._driver, '_allow_access')
        self.mock_object(self._driver, '_deny_access')

        existing_a_rules = [{'access_level': 'rw',
                             'access_type': 'ip',
                             'access_to': '10.0.0.2'},
                            {'access_level': 'rw',
                             'access_type': 'ip',
                             'access_to': '10.0.0.3'}]

        self._driver._fetch_existing_rule.return_value = existing_a_rules
        d_rule = self._driver._return_access_lists_difference(existing_a_rules,
                                                              [self.access4])

        a_rule = self._driver._return_access_lists_difference([self.access4],
                                                              existing_a_rules)
        self._driver.update_access(self._context, self.share,
                                   [self.access4], [], [])

        self.assertEqual(d_rule, existing_a_rules)
        self.assertEqual(a_rule, [self.access4])
        self.assertEqual(1, self._driver._allow_access.call_count)
        self.assertEqual(2, self._driver._deny_access.call_count)

    def test_extend_share(self):
        self.mock_object(self._driver, '_access_api')
        new_size = 3
        self._driver.extend_share(self.share, new_size)
        self.assertEqual(1, self._driver._access_api.call_count)

    def test_extend_share_negative(self):
        self.mock_object(self._driver, '_access_api')

        new_size = 3
        self._driver._access_api.return_value = False
        self.assertRaises(exception.ShareBackendException,
                          self._driver.extend_share,
                          self.share, new_size)

    def test_shrink_share(self):
        self.mock_object(self._driver, '_access_api')
        new_size = 3
        self._driver.shrink_share(self.share2, new_size)
        self.assertEqual(1, self._driver._access_api.call_count)

    def test_shrink_share_negative(self):
        self.mock_object(self._driver, '_access_api')

        new_size = 3
        self._driver._access_api.return_value = False
        self.assertRaises(exception.ShareBackendException,
                          self._driver.shrink_share,
                          self.share2, new_size)

    def test__get_access_pool_details(self):
        self.mock_object(self._driver, '_access_api')

        pool_details = []
        pool_details_dict = {}
        pool_details_dict['device_group_name'] = 'fake_pool'
        pool_details_dict['capacity'] = 10737418240
        pool_details_dict['used_size'] = 9663676416
        pool_details.append(pool_details_dict)

        pool_details_dict2 = {}
        pool_details_dict2['device_group_name'] = self.configuration.va_pool
        pool_details_dict2['capacity'] = 10737418240
        pool_details_dict2['used_size'] = 9663676416
        pool_details.append(pool_details_dict2)

        self._driver._access_api.return_value = pool_details
        total_space, free_space = self._driver._get_access_pool_details()
        self.assertEqual(10, total_space)
        self.assertEqual(1, free_space)

    def test__get_access_pool_details_negative(self):
        self.mock_object(self._driver, '_access_api')

        pool_details = []
        self._driver._access_api.return_value = pool_details
        self.assertRaises(exception.ShareBackendException,
                          self._driver._get_access_pool_details)

    def test__update_share_stats(self):
        self.mock_object(self._driver, '_authenticate_access')
        self.mock_object(self._driver, '_get_access_pool_details')

        self._driver._get_access_pool_details.return_value = (10, 9)
        self._driver._update_share_stats()
        data = {
            'share_backend_name': FAKE_BACKEND,
            'vendor_name': 'Veritas',
            'driver_version': '1.0',
            'storage_protocol': 'NFS',
            'total_capacity_gb': 10,
            'free_capacity_gb': 9,
            'reserved_percentage': 0,
            'reserved_snapshot_percentage': 0,
            'QoS_support': False,
            'create_share_from_snapshot_support': True,
            'driver_handles_share_servers': False,
            'filter_function': 'Disable',
            'goodness_function': 'Disable',
            'ipv4_support': True,
            'ipv6_support': False,
            'mount_snapshot_support': False,
            'pools': None,
            'qos': False,
            'replication_domain': 'Disable',
            'revert_to_snapshot_support': False,
            'share_group_stats': {'consistent_snapshot_support': None},
            'snapshot_support': True,
            'security_service_update_support': False,
        }

        self.assertEqual(data, self._driver._stats)

    def test__get_vip(self):
        self.mock_object(self._driver, '_get_access_ips')

        pool_list = []
        ip1 = {'isconsoleip': 1, 'type': 'Virtual',
               'status': 'ONLINE', 'ip': '1.1.1.2'}
        ip2 = {'isconsoleip': 0, 'type': 'Virtual',
               'status': 'ONLINE', 'ip': '1.1.1.4'}
        ip3 = {'isconsoleip': 0, 'type': 'Virtual',
               'status': 'OFFLINE', 'ip': '1.1.1.5'}
        ip4 = {'isconsoleip': 0, 'type': 'Physical',
               'status': 'OFFLINE', 'ip': '1.1.1.6'}

        pool_list = [ip1, ip2, ip3, ip4]

        self._driver._get_access_ips.return_value = pool_list
        vip = self._driver._get_vip()
        self.assertEqual('1.1.1.4', vip)

    def test__get_access_ips(self):
        self.mock_object(self._driver, '_access_api')
        ip_list = ['1.1.1.2', '1.1.2.3', '1.1.1.4']
        self._driver._access_api.return_value = ip_list
        ret_value = self._driver._get_access_ips(self._driver.session,
                                                 self._driver.host)
        self.assertEqual(ret_value, ip_list)

    def test__access_api(self):
        self.mock_object(requests, 'session')

        provider = '%s:%s' % (self._driver.host, self._driver._port)
        path = '/fake/path'
        input_data = {}
        mock_response = MockResponse()
        session = requests.session

        data = {'fake_key': 'fake_val'}
        json_data = json.dumps(data)

        session.request.return_value = mock_response
        ret_value = self._driver._access_api(session, provider, path,
                                             json.dumps(input_data), 'GET')

        self.assertEqual(json_data, ret_value)

    def test__access_api_ret_for_update_object(self):
        self.mock_object(requests, 'session')

        provider = '%s:%s' % (self._driver.host, self._driver._port)
        path = self._driver._update_object
        input_data = None
        mock_response = MockResponse()
        session = requests.session

        session.request.return_value = mock_response
        ret = self._driver._access_api(session, provider, path,
                                       input_data, 'GET')

        self.assertTrue(ret)

    def test__access_api_negative(self):
        session = self._driver.session
        provider = '%s:%s' % (self._driver.host, self._driver._port)
        path = '/fake/path'
        input_data = {}
        ret_value = self._driver._access_api(session, provider, path,
                                             json.dumps(input_data), 'GET')
        self.assertEqual(False, ret_value)

    def test__get_api(self):
        provider = '%s:%s' % (self._driver.host, self._driver._port)
        tail = '/fake/path'
        ret = self._driver._get_api(provider, tail)

        api_root = 'https://%s/api' % (provider)
        to_be_ret = api_root + tail
        self.assertEqual(to_be_ret, ret)

    def test__does_item_exist_at_va_backend(self):
        self.mock_object(self._driver, '_access_api')
        item_name = 'fake_item'
        path = '/fake/path'
        fake_item_list = [{'name': item_name}]
        self._driver._access_api.return_value = fake_item_list
        ret_value = self._driver._does_item_exist_at_va_backend(item_name,
                                                                path)
        self.assertTrue(ret_value)

    def test__does_item_exist_at_va_backend_negative(self):
        self.mock_object(self._driver, '_access_api')
        item_name = 'fake_item'
        path = '/fake/path'
        fake_item_list = [{'name': 'item2'}]
        self._driver._access_api.return_value = fake_item_list
        ret_value = self._driver._does_item_exist_at_va_backend(item_name,
                                                                path)
        self.assertEqual(False, ret_value)

    def test__fetch_existing_rule(self):
        self.mock_object(self._driver, '_access_api')
        fake_share = 'fake-share'
        fake_access_list = []
        list1 = []
        list1.append({
            'status': 'online',
            'name': '/vx/fake-share',
            'host_name': '10.0.0.1',
            'privilege': 'rw'
        })
        list1.append({
            'status': 'online',
            'name': '/vx/fake-share',
            'host_name': '10.0.0.2',
            'privilege': 'rw'
        })
        list1.append({
            'status': 'online',
            'name': '/vx/fake-share',
            'host_name': '10.0.0.3',
            'privilege': 'ro'
        })
        list1.append({
            'status': 'online',
            'name': '/vx/fake-share2',
            'host_name': '10.0.0.4',
            'privilege': 'rw'
        })

        fake_access_list.append({
            'shareType': 'NFS',
            'shares': list1
        })

        fake_access_list.append({
            'shareType': 'CIFS',
            'shares': []
        })

        ret_access_list = []
        ret_access_list.append({
            'access_to': '10.0.0.1',
            'access_level': 'rw',
            'access_type': 'ip'
        })

        ret_access_list.append({
            'access_to': '10.0.0.2',
            'access_level': 'rw',
            'access_type': 'ip'
        })

        ret_access_list.append({
            'access_to': '10.0.0.3',
            'access_level': 'ro',
            'access_type': 'ip'
        })

        self._driver._access_api.return_value = fake_access_list
        ret_value = self._driver._fetch_existing_rule(fake_share)
        self.assertEqual(ret_access_list, ret_value)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import six
from six.moves import http_client
import webob

from manila.api import common
from manila.api.openstack import wsgi
from manila import db
from manila import exception
from manila.i18n import _
from manila.share_group import share_group_types

SG_GRADUATION_VERSION = '2.55'


class ShareGroupTypeSpecsController(wsgi.Controller):
    """The share group type specs API controller for the OpenStack API."""

    resource_name = 'share_group_types_spec'

    def _get_group_specs(self, context, type_id):
        specs = db.share_group_type_specs_get(context, type_id)
        return {"group_specs": copy.deepcopy(specs)}

    def _assert_share_group_type_exists(self, context, type_id):
        try:
            share_group_types.get(context, type_id)
        except exception.NotFound as ex:
            raise webob.exc.HTTPNotFound(explanation=ex.msg)

    def _verify_group_specs(self, group_specs):

        def is_valid_string(v):
            return isinstance(v, six.string_types) and len(v) in range(1, 256)

        def is_valid_spec(k, v):
            valid_spec_key = is_valid_string(k)
            valid_type = is_valid_string(v) or isinstance(v, bool)
            return valid_spec_key and valid_type

        for k, v in group_specs.items():
            if is_valid_string(k) and isinstance(v, dict):
                self._verify_group_specs(v)
            elif not is_valid_spec(k, v):
                expl = _('Invalid extra_spec: %(key)s: %(value)s') % {
                    'key': k, 'value': v
                }
                raise webob.exc.HTTPBadRequest(explanation=expl)

    @wsgi.Controller.authorize('index')
    def _index(self, req, id):
        """Returns the list of group specs for a given share group type."""

        context = req.environ['manila.context']
        self._assert_share_group_type_exists(context, id)
        return self._get_group_specs(context, id)

    @wsgi.Controller.api_version('2.31', '2.54', experimental=True)
    def index(self, req, id):
        return self._index(req, id)

    @wsgi.Controller.api_version(SG_GRADUATION_VERSION)  # noqa
    def index(self, req, id):  # pylint: disable=function-redefined  # noqa F811
        return self._index(req, id)

    @wsgi.Controller.authorize('create')
    def _create(self, req, id, body=None):
        context = req.environ['manila.context']
        if not self.is_valid_body(body, 'group_specs'):
            raise webob.exc.HTTPBadRequest()

        self._assert_share_group_type_exists(context, id)
        specs = body['group_specs']
        self._verify_group_specs(specs)
        self._check_key_names(specs.keys())
        db.share_group_type_specs_update_or_create(context, id, specs)
        return body

    @wsgi.Controller.api_version('2.31', '2.54', experimental=True)
    def create(self, req, id, body=None):
        return self._create(req, id, body)

    @wsgi.Controller.api_version(SG_GRADUATION_VERSION)  # noqa
    def create(self, req, id, body=None):  # pylint: disable=function-redefined   # noqa F811
        return self._create(req, id, body)

    @wsgi.Controller.authorize('update')
    def _update(self, req, id, key, body=None):
        context = req.environ['manila.context']
        if not body:
            expl = _('Request body empty.')
            raise webob.exc.HTTPBadRequest(explanation=expl)
        self._assert_share_group_type_exists(context, id)
        if key not in body:
            expl = _('Request body and URI mismatch.')
            raise webob.exc.HTTPBadRequest(explanation=expl)
        if len(body) > 1:
            expl = _('Request body contains too many items.')
            raise webob.exc.HTTPBadRequest(explanation=expl)
        self._verify_group_specs(body)
        db.share_group_type_specs_update_or_create(context, id, body)
        return body

    @wsgi.Controller.api_version('2.31', '2.54', experimental=True)
    def update(self, req, id, key, body=None):
        return self._update(req, id, key, body)

    # pylint: disable=function-redefined
    @wsgi.Controller.api_version(SG_GRADUATION_VERSION)  # noqa
    def update(self, req, id, key, body=None):  # noqa F811
        return self._update(req, id, key, body)

    @wsgi.Controller.authorize('show')
    def _show(self, req, id, key):
        """Return a single group spec item."""
        context = req.environ['manila.context']
        self._assert_share_group_type_exists(context, id)
        specs = self._get_group_specs(context, id)
        if key in specs['group_specs']:
            return {key: specs['group_specs'][key]}
        else:
            raise webob.exc.HTTPNotFound()

    # pylint: enable=function-redefined
    @wsgi.Controller.api_version('2.31', '2.54', experimental=True)
    def show(self, req, id, key):
        return self._show(req, id, key)

    @wsgi.Controller.api_version(SG_GRADUATION_VERSION)  # noqa
    def show(self, req, id, key):  # pylint: disable=function-redefined   # noqa F811
        return self._show(req, id, key)

    @wsgi.Controller.authorize('delete')
    def _delete(self, req, id, key):
        """Deletes an existing group spec."""
        context = req.environ['manila.context']
        self._assert_share_group_type_exists(context, id)
        try:
            db.share_group_type_specs_delete(context, id, key)
        except exception.ShareGroupTypeSpecsNotFound as error:
            raise webob.exc.HTTPNotFound(explanation=error.msg)
        return webob.Response(status_int=http_client.NO_CONTENT)

    @wsgi.Controller.api_version('2.31', '2.54', experimental=True)
    def delete(self, req, id, key):
        return self._delete(req, id, key)

    @wsgi.Controller.api_version(SG_GRADUATION_VERSION)  # noqa
    def delete(self, req, id, key):  # pylint: disable=function-redefined   # noqa F811
        return self._delete(req, id, key)

    def _check_key_names(self, keys):
        if not common.validate_key_names(keys):
            expl = _('Key names can only contain alphanumeric characters, '
                     'underscores, periods, colons and hyphens.')

            raise webob.exc.HTTPBadRequest(explanation=expl)


def create_resource():
    return wsgi.Resource(ShareGroupTypeSpecsController())



# Copyright 2012 NetApp
# Copyright 2016 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
LVM Driver for shares.

"""

import ipaddress
import math
import os
import re

from oslo_config import cfg
from oslo_log import log
from oslo_utils import importutils
from oslo_utils import timeutils
import six

from manila import exception
from manila.i18n import _
from manila.share import driver
from manila.share.drivers import generic
from manila.share import utils as share_utils
from manila import utils

LOG = log.getLogger(__name__)

share_opts = [
    cfg.StrOpt('lvm_share_export_root',
               default='$state_path/mnt',
               help='Base folder where exported shares are located.'),
    cfg.ListOpt('lvm_share_export_ips',
                help='List of IPs to export shares belonging to the LVM '
                     'storage driver.'),
    cfg.IntOpt('lvm_share_mirrors',
               default=0,
               help='If set, create LVMs with multiple mirrors. Note that '
                    'this requires lvm_mirrors + 2 PVs with available space.'),
    cfg.StrOpt('lvm_share_volume_group',
               default='lvm-shares',
               help='Name for the VG that will contain exported shares.'),
    cfg.ListOpt('lvm_share_helpers',
                default=[
                    'CIFS=manila.share.drivers.helpers.CIFSHelperUserAccess',
                    'NFS=manila.share.drivers.helpers.NFSHelper',
                ],
                help='Specify list of share export helpers.'),
]

CONF = cfg.CONF
CONF.register_opts(share_opts)
CONF.register_opts(generic.share_opts)


class LVMMixin(driver.ExecuteMixin):
    def check_for_setup_error(self):
        """Returns an error if prerequisites aren't met."""
        out, err = self._execute('vgs', '--noheadings', '-o', 'name',
                                 run_as_root=True)
        volume_groups = out.split()
        if self.configuration.lvm_share_volume_group not in volume_groups:
            msg = (_("Share volume group %s doesn't exist.")
                   % self.configuration.lvm_share_volume_group)
            raise exception.InvalidParameterValue(err=msg)

        if not self.configuration.lvm_share_export_ips:
            msg = _("The option lvm_share_export_ips must be specified.")
            raise exception.InvalidParameterValue(err=msg)

    def _allocate_container(self, share):
        sizestr = '%sG' % share['size']
        cmd = ['lvcreate', '-L', sizestr, '-n', share['name'],
               self.configuration.lvm_share_volume_group]
        if self.configuration.lvm_share_mirrors:
            cmd += ['-m', self.configuration.lvm_share_mirrors, '--nosync']
            terras = int(sizestr[:-1]) / 1024.0
            if terras >= 1.5:
                rsize = int(2 ** math.ceil(math.log(terras) / math.log(2)))
                # NOTE(vish): Next power of two for region size. See:
                #             http://red.ht/U2BPOD
                cmd += ['-R', six.text_type(rsize)]

        self._try_execute(*cmd, run_as_root=True)
        device_name = self._get_local_path(share)
        self._execute('mkfs.%s' % self.configuration.share_volume_fstype,
                      device_name, run_as_root=True)

    def _extend_container(self, share, device_name, size):
        cmd = ['lvextend', '-L', '%sG' % size, '-r', device_name]
        self._try_execute(*cmd, run_as_root=True)

    def _deallocate_container(self, share_name):
        """Deletes a logical volume for share."""
        try:
            self._try_execute('lvremove', '-f', "%s/%s" %
                              (self.configuration.lvm_share_volume_group,
                               share_name), run_as_root=True)
        except exception.ProcessExecutionError as exc:
            err_pattern = re.compile(".*failed to find.*|.*not found.*",
                                     re.IGNORECASE)
            if not err_pattern.match(exc.stderr):
                LOG.exception("Error deleting volume")
                raise
            LOG.warning("Volume not found: %s", exc.stderr)

    def _create_snapshot(self, context, snapshot):
        """Creates a snapshot."""
        orig_lv_name = "%s/%s" % (self.configuration.lvm_share_volume_group,
                                  snapshot['share_name'])
        self._try_execute(
            'lvcreate', '-L', '%sG' % snapshot['share']['size'],
            '--name', snapshot['name'],
            '--snapshot', orig_lv_name, run_as_root=True)

        self._set_random_uuid_to_device(snapshot)

    def _set_random_uuid_to_device(self, share_or_snapshot):
        # NOTE(vponomaryov): 'tune2fs' is required to make
        # filesystem of share created from snapshot have
        # unique ID, in case of LVM volumes, by default,
        # it will have the same UUID as source volume. Closes #1645751
        # NOTE(gouthamr): Executing tune2fs -U only works on
        # a recently checked filesystem.
        # See: https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=857336
        device_path = self._get_local_path(share_or_snapshot)
        self._execute('e2fsck', '-y', '-f', device_path, run_as_root=True)
        self._execute(
            'tune2fs', '-U', 'random', device_path, run_as_root=True,
        )

    def create_snapshot(self, context, snapshot, share_server=None):
        self._create_snapshot(context, snapshot)

    def delete_snapshot(self, context, snapshot, share_server=None):
        """Deletes a snapshot."""
        self._deallocate_container(snapshot['name'])


class LVMShareDriver(LVMMixin, driver.ShareDriver):
    """Executes commands relating to Shares."""

    def __init__(self, *args, **kwargs):
        """Do initialization."""
        super(LVMShareDriver, self).__init__([False], *args, **kwargs)
        self.configuration.append_config_values(share_opts)
        self.configuration.append_config_values(generic.share_opts)
        self.configuration.share_mount_path = (
            self.configuration.lvm_share_export_root)
        self._helpers = None
        self.configured_ip_version = None
        self.backend_name = self.configuration.safe_get(
            'share_backend_name') or 'LVM'
        # Set of parameters used for compatibility with
        # Generic driver's helpers.
        self.share_server = {
            'instance_id': self.backend_name,
            'lock_name': 'manila_lvm',
        }
        self.share_server['public_addresses'] = (
            self.configuration.lvm_share_export_ips
        )
        self.ipv6_implemented = True

    def _ssh_exec_as_root(self, server, command, check_exit_code=True):
        kwargs = {}
        if 'sudo' in command:
            kwargs['run_as_root'] = True
            command.remove('sudo')
        kwargs['check_exit_code'] = check_exit_code
        return self._execute(*command, **kwargs)

    def do_setup(self, context):
        """Any initialization the volume driver does while starting."""
        super(LVMShareDriver, self).do_setup(context)
        self._setup_helpers()

    def _setup_helpers(self):
        """Initializes protocol-specific NAS drivers."""
        self._helpers = {}
        for helper_str in self.configuration.lvm_share_helpers:
            share_proto, _, import_str = helper_str.partition('=')
            helper = importutils.import_class(import_str)
            # TODO(rushiagr): better way to handle configuration
            #                 instead of just passing to the helper
            self._helpers[share_proto.upper()] = helper(
                self._execute, self._ssh_exec_as_root, self.configuration)

    def _get_local_path(self, share):
        # The escape characters are expected by the device mapper.
        escaped_group = (
            self.configuration.lvm_share_volume_group.replace('-', '--'))
        escaped_name = share['name'].replace('-', '--')
        return "/dev/mapper/%s-%s" % (escaped_group, escaped_name)

    def _update_share_stats(self):
        """Retrieve stats info from share volume group."""
        data = {
            'share_backend_name': self.backend_name,
            'storage_protocol': 'NFS_CIFS',
            'reserved_percentage':
                self.configuration.reserved_share_percentage,
            'reserved_snapshot_percentage':
                (self.configuration.reserved_share_from_snapshot_percentage
                    or self.configuration.reserved_share_percentage),
            'snapshot_support': True,
            'create_share_from_snapshot_support': True,
            'revert_to_snapshot_support': True,
            'mount_snapshot_support': True,
            'driver_name': 'LVMShareDriver',
            'pools': self.get_share_server_pools(),
        }
        super(LVMShareDriver, self)._update_share_stats(data)

    def get_share_server_pools(self, share_server=None):
        out, err = self._execute('vgs',
                                 self.configuration.lvm_share_volume_group,
                                 '--rows', '--units', 'g',
                                 run_as_root=True)
        total_size = re.findall(r"VSize\s[0-9.]+g", out)[0][6:-1]
        free_size = re.findall(r"VFree\s[0-9.]+g", out)[0][6:-1]
        return [{
            'pool_name': 'lvm-single-pool',
            'total_capacity_gb': float(total_size),
            'free_capacity_gb': float(free_size),
            'reserved_percentage': 0,
            'reserved_snapshot_percentage': 0,
        }, ]

    def create_share(self, context, share, share_server=None):
        self._allocate_container(share)
        # create file system
        device_name = self._get_local_path(share)
        location = self._get_helper(share).create_exports(
            self.share_server, share['name'])
        self._mount_device(share, device_name)
        return location

    def create_share_from_snapshot(self, context, share, snapshot,
                                   share_server=None, parent_share=None):
        """Is called to create share from snapshot."""
        self._allocate_container(share)
        snapshot_device_name = self._get_local_path(snapshot)
        share_device_name = self._get_local_path(share)
        self._set_random_uuid_to_device(share)
        self._copy_volume(
            snapshot_device_name, share_device_name, share['size'])
        location = self._get_helper(share).create_exports(
            self.share_server, share['name'])
        self._mount_device(share, share_device_name)
        return location

    def delete_share(self, context, share, share_server=None):
        self._unmount_device(share, raise_if_missing=False,
                             retry_busy_device=True)
        self._delete_share(context, share)
        self._deallocate_container(share['name'])

    def _unmount_device(self, share_or_snapshot, raise_if_missing=True,
                        retry_busy_device=False):
        """Unmount the filesystem of a share or snapshot LV."""
        mount_path = self._get_mount_path(share_or_snapshot)
        if os.path.exists(mount_path):

            retries = 10 if retry_busy_device else 1

            @utils.retry(retry_param=exception.ShareBusyException,
                         retries=retries)
            def _unmount_device_with_retry():
                try:
                    self._execute('umount', '-f', mount_path, run_as_root=True)
                except exception.ProcessExecutionError as exc:
                    if 'is busy' in exc.stderr.lower():
                        raise exception.ShareBusyException(
                            reason=share_or_snapshot['name'])
                    elif 'not mounted' in exc.stderr.lower():
                        if raise_if_missing:
                            LOG.error('Unable to find device: %s', exc)
                            raise
                    else:
                        LOG.error('Unable to umount: %s', exc)
                        raise

            _unmount_device_with_retry()
            # remove dir
            self._execute('rmdir', mount_path, run_as_root=True)

    def ensure_shares(self, context, shares):
        updates = {}
        for share in shares:
            updates[share['id']] = {
                'export_locations': self.ensure_share(context, share)}
        return updates

    def ensure_share(self, ctx, share, share_server=None):
        """Ensure that storage are mounted and exported."""
        device_name = self._get_local_path(share)
        self._mount_device(share, device_name)
        return self._get_helper(share).create_exports(
            self.share_server, share['name'], recreate=True)

    def _delete_share(self, ctx, share):
        """Delete a share."""
        try:
            self._get_helper(share).remove_exports(
                self.share_server, share['name'])
        except exception.ProcessExecutionError:
            LOG.warning("Can't remove share %r", share['id'])
        except exception.InvalidShare as exc:
            LOG.warning(exc)

    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        """Update access rules for given share.

        This driver has two different behaviors according to parameters:
        1. Recovery after error - 'access_rules' contains all access_rules,
        'add_rules' and 'delete_rules' shall be empty. Previously existing
        access rules are cleared and then added back according
        to 'access_rules'.

        2. Adding/Deleting of several access rules - 'access_rules' contains
        all access_rules, 'add_rules' and 'delete_rules' contain rules which
        should be added/deleted. Rules in 'access_rules' are ignored and
        only rules from 'add_rules' and 'delete_rules' are applied.

        :param context: Current context
        :param share: Share model with share data.
        :param access_rules: All access rules for given share
        :param add_rules: Empty List or List of access rules which should be
               added. access_rules already contains these rules.
        :param delete_rules: Empty List or List of access rules which should be
               removed. access_rules doesn't contain these rules.
        :param share_server: None or Share server model
        """
        self._get_helper(share).update_access(self.share_server,
                                              share['name'], access_rules,
                                              add_rules=add_rules,
                                              delete_rules=delete_rules)

    def _get_helper(self, share):
        if share['share_proto'].lower().startswith('nfs'):
            return self._helpers['NFS']
        elif share['share_proto'].lower().startswith('cifs'):
            return self._helpers['CIFS']
        else:
            raise exception.InvalidShare(reason='Wrong share protocol')

    def _mount_device(self, share_or_snapshot, device_name):
        """Mount LV for share or snapshot and ignore if already mounted."""
        mount_path = self._get_mount_path(share_or_snapshot)
        self._execute('mkdir', '-p', mount_path)
        try:
            self._execute('mount', device_name, mount_path,
                          run_as_root=True, check_exit_code=True)
            self._execute('chmod', '777', mount_path,
                          run_as_root=True, check_exit_code=True)
        except exception.ProcessExecutionError:
            out, err = self._execute('mount', '-l', run_as_root=True)
            if device_name in out:
                LOG.warning("%s is already mounted", device_name)
            else:
                raise
        return mount_path

    def _get_mount_path(self, share_or_snapshot):
        """Returns path where share or snapshot is mounted."""
        return os.path.join(self.configuration.share_mount_path,
                            share_or_snapshot['name'])

    def _copy_volume(self, srcstr, deststr, size_in_g):
        # Use O_DIRECT to avoid thrashing the system buffer cache
        extra_flags = ['iflag=direct', 'oflag=direct']

        # Check whether O_DIRECT is supported
        try:
            self._execute('dd', 'count=0', 'if=%s' % srcstr, 'of=%s' % deststr,
                          *extra_flags, run_as_root=True)
        except exception.ProcessExecutionError:
            extra_flags = []

        # Perform the copy
        self._execute('dd', 'if=%s' % srcstr, 'of=%s' % deststr,
                      'count=%d' % (size_in_g * 1024), 'bs=1M',
                      *extra_flags, run_as_root=True)

    def extend_share(self, share, new_size, share_server=None):
        device_name = self._get_local_path(share)
        self._extend_container(share, device_name, new_size)

    def revert_to_snapshot(self, context, snapshot, share_access_rules,
                           snapshot_access_rules, share_server=None):
        share = snapshot['share']
        # Temporarily remove all access rules
        self._get_helper(share).update_access(self.share_server,
                                              snapshot['name'], [], [], [])
        self._get_helper(share).update_access(self.share_server,
                                              share['name'], [], [], [])
        # Unmount the snapshot filesystem
        self._unmount_device(snapshot)
        # Unmount the share filesystem
        self._unmount_device(share)
        # Merge the snapshot LV back into the share, reverting it
        snap_lv_name = "%s/%s" % (self.configuration.lvm_share_volume_group,
                                  snapshot['name'])
        self._execute('lvconvert', '--merge', snap_lv_name, run_as_root=True)

        # Now recreate the snapshot that was destroyed by the merge
        self._create_snapshot(context, snapshot)
        # At this point we can mount the share again
        device_name = self._get_local_path(share)
        self._mount_device(share, device_name)
        # Also remount the snapshot
        device_name = self._get_local_path(snapshot)
        self._mount_device(snapshot, device_name)
        # Lastly we add all the access rules back
        self._get_helper(share).update_access(self.share_server,
                                              share['name'],
                                              share_access_rules,
                                              [], [])
        snapshot_access_rules, __, __ = share_utils.change_rules_to_readonly(
            snapshot_access_rules, [], [])
        self._get_helper(share).update_access(self.share_server,
                                              snapshot['name'],
                                              snapshot_access_rules,
                                              [], [])

    def create_snapshot(self, context, snapshot, share_server=None):
        self._create_snapshot(context, snapshot)

        device_name = self._get_local_path(snapshot)
        self._mount_device(snapshot, device_name)

        helper = self._get_helper(snapshot['share'])
        exports = helper.create_exports(self.share_server, snapshot['name'])

        return {'export_locations': exports}

    def delete_snapshot(self, context, snapshot, share_server=None):
        self._unmount_device(snapshot, raise_if_missing=False)

        super(LVMShareDriver, self).delete_snapshot(context, snapshot,
                                                    share_server)

    def get_configured_ip_versions(self):
        if self.configured_ip_version is None:
            try:
                self.configured_ip_version = []
                for ip in self.configuration.lvm_share_export_ips:
                    self.configured_ip_version.append(
                        ipaddress.ip_address(six.text_type(ip)).version)
            except Exception:
                message = (_("Invalid 'lvm_share_export_ips' option supplied "
                             "%s.") % self.configuration.lvm_share_export_ips)
                raise exception.InvalidInput(reason=message)
        return self.configured_ip_version

    def snapshot_update_access(self, context, snapshot, access_rules,
                               add_rules, delete_rules, share_server=None):
        """Update access rules for given snapshot.

        This driver has two different behaviors according to parameters:
        1. Recovery after error - 'access_rules' contains all access_rules,
        'add_rules' and 'delete_rules' shall be empty. Previously existing
        access rules are cleared and then added back according
        to 'access_rules'.

        2. Adding/Deleting of several access rules - 'access_rules' contains
        all access_rules, 'add_rules' and 'delete_rules' contain rules which
        should be added/deleted. Rules in 'access_rules' are ignored and
        only rules from 'add_rules' and 'delete_rules' are applied.

        :param context: Current context
        :param snapshot: Snapshot model with snapshot data.
        :param access_rules: All access rules for given snapshot
        :param add_rules: Empty List or List of access rules which should be
               added. access_rules already contains these rules.
        :param delete_rules: Empty List or List of access rules which should be
               removed. access_rules doesn't contain these rules.
        :param share_server: None or Share server model
        """
        helper = self._get_helper(snapshot['share'])
        access_rules, add_rules, delete_rules = (
            share_utils.change_rules_to_readonly(
                access_rules, add_rules, delete_rules)
        )

        helper.update_access(self.share_server,
                             snapshot['name'], access_rules,
                             add_rules=add_rules, delete_rules=delete_rules)

    def update_share_usage_size(self, context, shares):
        updated_shares = []
        out, err = self._execute(
            'df', '-l', '--output=target,used',
            '--block-size=g')
        gathered_at = timeutils.utcnow()

        for share in shares:
            try:
                mount_path = self._get_mount_path(share)
                if os.path.exists(mount_path):
                    used_size = (re.findall(
                        mount_path + r"\s*[0-9.]+G", out)[0].
                        split(' ')[-1][:-1])
                    updated_shares.append({'id': share['id'],
                                           'used_size': used_size,
                                           'gathered_at': gathered_at})
                else:
                    raise exception.NotFound(
                        _("Share mount path %s could not be "
                          "found.") % mount_path)
            except Exception:
                LOG.exception("Failed to gather 'used_size' for share %s.",
                              share['id'])

        return updated_shares

    def get_backend_info(self, context):
        return {
            'export_ips': ','.join(self.share_server['public_addresses']),
            'db_version': share_utils.get_recent_db_migration_id(),
        }



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Implementation of paginate query."""

from manila import exception
import sqlalchemy


def paginate_query(query, model, limit, sort_key='created_at',
                   sort_dir='desc', offset=None):
    """Returns a query with sorting / pagination criteria added.

    :param query: the query object to which we should add paging/sorting
    :param model: the ORM model class
    :param limit: maximum number of items to return
    :param sort_key: attributes by which results should be sorted, default is
                      created_at
    :param sort_dir: direction in which results should be sorted (asc, desc)
    :param offset: the number of items to skip from the marker or from the
                    first element.

    :rtype: sqlalchemy.orm.query.Query
    :return: The query with sorting/pagination added.
    """

    try:
        sort_key_attr = getattr(model, sort_key)
    except AttributeError:
        raise exception.InvalidInput(reason='Invalid sort key %s' % sort_key)
    if sort_dir == 'desc':
        query = query.order_by(sqlalchemy.desc(sort_key_attr))
    else:
        query = query.order_by(sqlalchemy.asc(sort_key_attr))

    if limit is not None:
        query = query.limit(limit)

    if offset:
        query = query.offset(offset)

    return query
